from pycobertura import Cobertura
from pycobertura.filesystem import filesystem_factory
from pycobertura.utils import get_dir_from_file_path


def make_cobertura(xml='tests/cobertura.xml', source=None, **kwargs):
    if not source:
        source = get_dir_from_file_path(xml)
    source_filesystem = filesystem_factory(source, **kwargs)
    cobertura = Cobertura(xml, filesystem=source_filesystem)
    return cobertura



import subprocess
from unittest.mock import patch, MagicMock


FIRST_PYCOBERTURA_COMMIT_SHA = "d1fe88da6b18340762b24bb1f89067a3439c4041"


def test_filesystem_directory__file_not_found():
    from pycobertura.filesystem import DirectoryFileSystem

    fs = DirectoryFileSystem('foo/bar/baz')

    expected_filepaths = {
        'Main.java': 'foo/bar/baz/Main.java',
        'search/BinarySearch.java': 'foo/bar/baz/search/BinarySearch.java',
        'search/ISortedArraySearch.java': 'foo/bar/baz/search/ISortedArraySearch.java',
        'search/LinearSearch.java': 'foo/bar/baz/search/LinearSearch.java',
    }

    for filename in expected_filepaths:
        try:
            with fs.open(filename) as f:
                pass
        except DirectoryFileSystem.FileNotFound as fnf:
            assert fnf.path == expected_filepaths[filename]


def test_filesystem_directory__returns_fileobject():
    from pycobertura.filesystem import DirectoryFileSystem

    fs = DirectoryFileSystem('tests/dummy')

    expected_filepaths = {
        'dummy/dummy.py': 'dummy/dummy/dummy.py',
    }

    for filename in expected_filepaths:
        with fs.open(filename) as f:
            assert hasattr(f, 'read')


def test_filesystem_directory__with_source_prefix():
    from pycobertura.filesystem import DirectoryFileSystem

    fs = DirectoryFileSystem(
        'tests/',
        source_prefix='dummy'  # should resolve to tests/dummy/
    )

    expected_filepaths = {
        'dummy/dummy.py': 'dummy/dummy/dummy.py',
    }

    for filename in expected_filepaths:
        with fs.open(filename) as f:
            assert hasattr(f, 'read')


def test_filesystem_zip__file_not_found():
    from pycobertura.filesystem import ZipFileSystem

    fs = ZipFileSystem("tests/dummy/dummy.zip")

    dummy_source_file = 'foo/non-existent-file.py'
    try:
        with fs.open(dummy_source_file) as f:
            pass
    except ZipFileSystem.FileNotFound as fnf:
        assert fnf.path == dummy_source_file


def test_filesystem_zip__returns_fileobject():
    from pycobertura.filesystem import ZipFileSystem

    fs = ZipFileSystem("tests/dummy/dummy.zip")

    source_files_in_zip = [
        'dummy/dummy.py',
        'dummy/dummy2.py',
    ]

    for source_file in source_files_in_zip:
        with fs.open(source_file) as f:
            assert hasattr(f, 'read')


def test_filesystem_zip__with_source_prefix():
    from pycobertura.filesystem import ZipFileSystem

    fs = ZipFileSystem(
        "tests/dummy/dummy-with-prefix.zip",  # code zipped as dummy-with-prefix/dummy/dummy.py
        source_prefix="dummy-with-prefix",
    )

    source_files_in_zip = [
        'dummy/dummy.py',
        'dummy/dummy2.py',
    ]

    for source_file in source_files_in_zip:
        with fs.open(source_file) as f:
            assert hasattr(f, 'read')


def test_filesystem_git():
    import pycobertura.filesystem as fsm

    branch, folder, filename = "master", "tests/dummy", "test-file"

    with patch.object(fsm, "subprocess") as subprocess_mock:
        subprocess_mock.check_output = MagicMock(return_value=b"<file-content>")

        fs = fsm.GitFileSystem(folder, branch)

        with fs.open(filename) as f:
            assert hasattr(f, 'read')

        expected_git_filename = "master:tests/dummy/test-file"
        git_filename = fs.real_filename(filename)
        assert git_filename == expected_git_filename

        expected_command = ["git", "--no-pager", "show", git_filename]
        subprocess_mock.check_output.assert_called_with(expected_command, cwd=folder)


def test_filesystem_git_integration():
    from pycobertura.filesystem import GitFileSystem

    fs = GitFileSystem(".", FIRST_PYCOBERTURA_COMMIT_SHA)

    # Files included in pycobertura's first commit.
    source_files = [
        'README.md',
        '.gitignore',
    ]

    for source_file in source_files:
        with fs.open(source_file) as f:
            assert hasattr(f, 'read')


def test_filesystem_git_integration__not_found():
    from pycobertura.filesystem import GitFileSystem

    fs = GitFileSystem(".", FIRST_PYCOBERTURA_COMMIT_SHA)

    dummy_source_file = "CHANGES.md"

    try:
        with fs.open(dummy_source_file) as f:
            pass
    except GitFileSystem.FileNotFound as fnf:
        assert fnf.path == fs.real_filename(dummy_source_file)


def test_filesystem_git__git_not_found():
    import pycobertura.filesystem as fsm

    branch, folder, filename = "master", "tests/dummy", "test-file"

    error = subprocess.CalledProcessError
    with patch.object(fsm, "subprocess") as subprocess_mock:
        subprocess_mock.check_output = MagicMock(side_effect=OSError)
        subprocess_mock.CalledProcessError = error

        try:
            fs = fsm.GitFileSystem(folder, branch)
        except ValueError as e:
            assert folder in str(e)


def test_filesystem_git_integration():
    from pycobertura.filesystem import GitFileSystem

    fs = GitFileSystem(".", FIRST_PYCOBERTURA_COMMIT_SHA)

    # Files included in pycobertura's first commit.
    source_files = [
        "README.md",
        ".gitignore",
    ]

    for source_file in source_files:
        with fs.open(source_file) as f:
            assert hasattr(f, "read")


def test_filesystem_git_has_file_integration():
    from pycobertura.filesystem import GitFileSystem

    fs = GitFileSystem(".", FIRST_PYCOBERTURA_COMMIT_SHA)

    # Files included in pycobertura's first commit.
    source_files = [
        "README.md",
        ".gitignore",
    ]

    for source_file in source_files:
        assert fs.has_file(source_file), source_file


def test_filesystem_git_integration__not_found():
    from pycobertura.filesystem import GitFileSystem

    fs = GitFileSystem(".", FIRST_PYCOBERTURA_COMMIT_SHA)

    dummy_source_file = "CHANGES.md"

    try:
        with fs.open(dummy_source_file) as f:
            pass
    except GitFileSystem.FileNotFound as fnf:
        assert fnf.path == fs.real_filename(dummy_source_file)


def test_filesystem_git_has_file_integration__not_found():
    from pycobertura.filesystem import GitFileSystem

    fs = GitFileSystem(".", FIRST_PYCOBERTURA_COMMIT_SHA)

    dummy_source_file = "CHANGES.md"

    assert not fs.has_file(dummy_source_file)


def test_filesystem_factory():
    from pycobertura.filesystem import filesystem_factory, GitFileSystem

    fs = filesystem_factory(source=".", ref=FIRST_PYCOBERTURA_COMMIT_SHA)
    assert isinstance(fs, GitFileSystem)



import difflib
import os
import re
import fnmatch

from functools import partial

ANSI_ESCAPE_CODES = {
    "green": "\x1b[32m",
    "red": "\x1b[31m",
    "reset": "\x1b[39m",
}


# Recipe from https://github.com/ActiveState/
# code/recipes/Python/577452_memoize_decorator_instance/recipe-577452.py
class memoize:
    """cache the return value of a method

    This class is meant to be used as a decorator of methods. The return value
    from a given method invocation will be cached on the instance whose method
    was invoked. All arguments passed to a method decorated with memoize must
    be hashable.

    If a memoized method is invoked directly on its class the result will not
    be cached. Instead the method will be invoked like a static method:
    class Obj:
        @memoize
        def add_to(self, arg):
            return self + arg
    Obj.add_to(1) # not enough arguments
    Obj.add_to(1, 2) # returns 3, result is not cached
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, args[1:], frozenset(kw.items()))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res


def colorize(text, color):
    color_code = ANSI_ESCAPE_CODES[color]
    return f'{color_code}{text}{ANSI_ESCAPE_CODES["reset"]}'


def red(text):
    return colorize(text, "red")


def green(text):
    return colorize(text, "green")


def rangify(number_list):
    """Assumes the list is sorted."""
    if not number_list:
        return number_list

    ranges = []

    range_start = prev_num = number_list[0]
    for num in number_list[1:]:
        if num != (prev_num + 1):
            ranges.append((range_start, prev_num))
            range_start = num
        prev_num = num

    ranges.append((range_start, prev_num))
    return ranges


def stringify(number_list):
    """Assumes the list is sorted."""
    rangified_list = rangify(number_list)
    stringified_list = [
        f"{line_start}" if line_start == line_stop else f"{line_start}-{line_stop}"
        for line_start, line_stop in rangified_list
    ]
    return ", ".join(stringified_list)


def extrapolate_coverage(lines_w_status):
    """
    Given the following input:

    >>> lines_w_status = [
        (1, True),
        (4, True),
        (7, False),
        (9, False),
    ]

    Return expanded lines with their extrapolated line status.

    >>> extrapolate_coverage(lines_w_status) == [
        (1, True),
        (2, True),
        (3, True),
        (4, True),
        (5, None),
        (6, None),
        (7, False),
        (8, False),
        (9, False),
    ]

    """
    lines = []

    prev_lineno = 0
    prev_status = True
    for lineno, status in lines_w_status:
        while (lineno - prev_lineno) > 1:
            prev_lineno += 1
            if prev_status is status:
                lines.append((prev_lineno, status))
            else:
                lines.append((prev_lineno, None))
        lines.append((lineno, status))
        prev_lineno = lineno
        prev_status = status

    return lines


def reconcile_lines(lines1, lines2):
    """
    Return a dict `{lineno1: lineno2}` which reconciles line numbers `lineno1`
    of list `lines1` to line numbers `lineno2` of list `lines2`. Only lines
    that are common in both sets are present in the dict, lines unique to one
    of the sets are omitted.
    """
    differ = difflib.Differ()
    diff = differ.compare(lines1, lines2)

    SAME = "  "
    ADDED = "+ "
    REMOVED = "- "
    INFO = "? "

    lineno_map = {}  # {lineno1: lineno2, ...}
    lineno1_offset = 0
    lineno2 = 1

    for diffline in diff:
        if diffline.startswith(INFO):
            continue

        if diffline.startswith(SAME):
            lineno1 = lineno2 + lineno1_offset
            lineno_map[lineno1] = lineno2

        elif diffline.startswith(ADDED):
            lineno1_offset -= 1

        elif diffline.startswith(REMOVED):
            lineno1_offset += 1
            continue

        lineno2 += 1

    return lineno_map


def hunkify_lines(lines, context=3):
    """
    Return a list of line hunks given a list of lines `lines`. The number of
    context lines can be control with `context` which will return line hunks
    surrounded with `context` lines before and after the code change.
    """
    # Find contiguous line changes
    ranges = []
    range_start = None
    for i, line in enumerate(lines):
        if line.status is not None:
            if range_start is None:
                range_start = i
                continue
        elif range_start is not None:
            range_stop = i
            ranges.append((range_start, range_stop))
            range_start = None
    else:
        # Append the last range
        if range_start is not None:
            range_stop = i
            ranges.append((range_start, range_stop))

    # add context
    ranges_w_context = []
    for range_start, range_stop in ranges:
        range_start = range_start - context
        range_start = range_start if range_start >= 0 else 0
        range_stop = range_stop + context
        ranges_w_context.append((range_start, range_stop))

    # merge overlapping hunks
    merged_ranges = ranges_w_context[:1]
    for range_start, range_stop in ranges_w_context[1:]:
        prev_start, prev_stop = merged_ranges[-1]
        if range_start <= prev_stop:
            range_start = prev_start
            merged_ranges[-1] = (range_start, range_stop)
        else:
            merged_ranges.append((range_start, range_stop))

    # build final hunks
    hunks = []
    for range_start, range_stop in merged_ranges:
        hunk = lines[range_start:range_stop]
        hunks.append(hunk)

    return hunks


def get_dir_from_file_path(file_path):
    return os.path.dirname(file_path) or "."


def get_non_empty_non_commented_lines_from_file_in_ascii(file_path, comment_character):
    with open(file_path, "rb") as f:  # read in binary (more secure)
        result = [
            line.decode("ascii").strip()
            for line in f.readlines()
            if not line.decode("ascii").startswith(comment_character)
        ]
        return [res for res in result if res != ""]


def get_filenames_that_do_not_match_regex(
    filenames, regex_param, comment_character="#"
):
    if os.path.isfile(regex_param):
        ignore_patterns = get_non_empty_non_commented_lines_from_file_in_ascii(
            regex_param, comment_character
        )
        remove_filenames = [
            filename
            for igp in ignore_patterns
            for filename in fnmatch.filter(filenames, igp)
        ]
    else:
        remove_filenames = list(filter(re.compile(regex_param).match, filenames))
    return [fname for fname in filenames if fname not in remove_filenames]



import codecs
import os
import io
import zipfile
import subprocess
import shlex

from contextlib import contextmanager


class FileSystem:
    class FileNotFound(Exception):
        def __init__(self, path):
            super(self.__class__, self).__init__(path)
            self.path = path


class DirectoryFileSystem(FileSystem):
    def __init__(self, source_dir, source_prefix=None):
        self.source_dir = source_dir
        self.source_prefix = source_prefix

    def real_filename(self, filename):
        if self.source_prefix is not None:
            filename = os.path.join(self.source_prefix, filename)
        return os.path.join(self.source_dir, filename)

    def has_file(self, filename):
        # FIXME: make this O(1)
        filename = self.real_filename(filename)
        return os.path.isfile(filename)

    @contextmanager
    def open(self, filename):
        """
        Yield a file-like object for file `filename`.

        This function is a context manager.
        """
        filename = self.real_filename(filename)

        if not os.path.exists(filename):
            raise self.FileNotFound(filename)

        with codecs.open(filename, encoding="utf-8") as f:
            yield f


class ZipFileSystem(FileSystem):
    def __init__(self, zip_file, source_prefix=None):
        self.zipfile = zipfile.ZipFile(zip_file)
        self.source_prefix = source_prefix

    def real_filename(self, filename):
        if self.source_prefix is not None:
            filename = os.path.join(self.source_prefix, filename)

        return filename

    def has_file(self, filename):
        # FIXME: make this O(1)
        return self.real_filename(filename) in self.zipfile.namelist()

    @contextmanager
    def open(self, filename):
        filename = self.real_filename(filename)

        try:
            f = self.zipfile.open(filename)
            yield f
            f.close()
        except KeyError:
            raise self.FileNotFound(filename)


class GitFileSystem(FileSystem):
    def __init__(self, repo_folder, ref):
        self.repository = repo_folder
        self.ref = ref
        self.repository_root = self._get_root_path(repo_folder)
        # the report may have been collected in a subfolder of the repository
        # root. Each file path shall thus be completed by the prefix.
        self.prefix = self.repository.replace(self.repository_root, "").lstrip("/")

    def real_filename(self, filename):
        prefix = "{}/".format(self.prefix) if self.prefix else ""
        return "{ref}:{prefix}{filename}".format(
            prefix=prefix, ref=self.ref, filename=filename
        )

    def has_file(self, filename):
        command = "git --no-pager show {}".format(self.real_filename(filename))
        return_code = subprocess.call(
            command,
            cwd=self.repository,
            shell=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return not bool(return_code)

    def _get_root_path(self, repository_folder):
        command = "git rev-parse --show-toplevel"
        command_tokens = shlex.split(command)
        try:
            output = subprocess.check_output(command_tokens, cwd=repository_folder)
        except (OSError, subprocess.CalledProcessError):
            raise ValueError(
                "The folder {} is not "
                "a valid git repository.".format(repository_folder)
            )

        return output.decode("utf-8").rstrip()

    @contextmanager
    def open(self, filename):
        """
        Yield a file-like object for file `filename`.

        This function is a context manager.
        """
        filename = self.real_filename(filename)

        command = "git --no-pager show {}".format(filename)
        command_tokens = shlex.split(command)

        try:
            output = subprocess.check_output(command_tokens, cwd=self.repository)
        except (OSError, subprocess.CalledProcessError):
            raise self.FileNotFound(filename)

        output = output.decode("utf-8").rstrip()
        yield io.StringIO(output)


def filesystem_factory(source, source_prefix=None, ref=None):
    """
    The argument `source` is the location of the source code provided as a
    directory path or a file object zip archive containing the source code.

    The optional argument `source` is the location of the source code provided
    as a directory path or a file object zip archive containing the source code.

    The optional argument `ref` will be taken into account when instantiating a
    GitFileSystem, and it shall be a branch name, a commit ID or a git ref ID.
    """
    if zipfile.is_zipfile(source):
        return ZipFileSystem(source, source_prefix=source_prefix)

    if ref:
        return GitFileSystem(source, ref)

    return DirectoryFileSystem(source, source_prefix=source_prefix)



import lxml.etree as ET

from collections import namedtuple

from pycobertura.utils import (
    extrapolate_coverage,
    reconcile_lines,
    hunkify_lines,
    get_filenames_that_do_not_match_regex,
    memoize,
)

try:
    basestring
except NameError:  # pragma: no cover
    # PY3 basestring
    basestring = (str, bytes)


class Line(namedtuple("Line", ["number", "source", "status", "reason"])):
    """
    A namedtuple object representing a line of source code.

    The namedtuple has the following attributes:
    `number`: line number in the source code
    `source`: actual source code of line
    `status`: True (covered), False (uncovered) or None (coverage unchanged)
    `reason`: If `Line.status` is not `None` the possible values may be
        `"line-edit"`, `"cov-up"` or `"cov-down"`. Otherwise `None`.
    """


class Cobertura:
    """
    An XML Cobertura parser.
    """

    class InvalidCoverageReport(Exception):
        pass

    class MissingFileSystem(Exception):
        pass

    def __init__(self, report, filesystem=None):
        """
        Initialize a Cobertura report given a coverage report `report` that is
        an XML file in the Cobertura format. It can represented as either:

            - a file object
            - a file path
            - an XML string

        The optional keyword argument `filesystem` describes how to retrieve the
        source files referenced in the report. Please check the
        `pycobertura.filesystem` module to learn more about filesystems.
        """
        for load_func in [
            self._load_from_file,
            self._load_from_string,
        ]:
            try:
                self.xml = load_func(report)
                break
            except BaseException:
                pass
        else:
            raise self.InvalidCoverageReport("Invalid coverage file: {}".format(report))

        self.filesystem = filesystem
        self.report = report

        self._class_elements_by_file_name = self._make_class_elements_by_filename()

    def _make_class_elements_by_filename(self):
        result = {}
        for elem in self.xml.xpath("./packages//class"):
            filename = elem.attrib["filename"]
            result.setdefault(filename, []).append(elem)

        return result

    def __eq__(self, other):
        return self.report and other.report and self.report == other.report

    def _load_from_file(self, report_file):
        return ET.parse(report_file).getroot()

    def _load_from_string(self, s):
        return ET.fromstring(s)

    @memoize
    def _get_lines_by_filename(self, filename):
        classElements = self._class_elements_by_file_name[filename]
        return [
            line
            for classElement in classElements
            for line in classElement.xpath("./lines/line")
        ]

    @property
    def version(self):
        """Return the version number of the coverage report."""
        return self.xml.get("version")

    def line_rate(self, filename=None):
        """
        Return the global line rate of the coverage report. If the
        `filename` file is given, return the line rate of the file.
        """

        if filename is None:
            return float(self.xml.get("line-rate"))

        elements = self._class_elements_by_file_name[filename]
        if len(elements) == 1:
            return float(elements[0].get("line-rate"))
        else:
            total = self.total_statements(filename)
            return float(self.total_hits(filename) / total) if total != 0 else 0

    def branch_rate(self, filename=None):
        """
        Return the global branch rate of the coverage report. If the
        `filename` file is given, return the branch rate of the file.
        """
        if filename is None:
            return float(self.xml.get("branch-rate"))
        else:
            classElement = self._class_elements_by_file_name[filename][0]
            return float(classElement.get("branch-rate"))

    @memoize
    def missed_statements(self, filename):
        """
        Return a list of uncovered line numbers for each of the missed
        statements found for the file `filename`.
        """
        classElements = self._class_elements_by_file_name[filename]
        return [
            int(line.get("number"))
            for classElement in classElements
            for line in classElement.xpath("./lines/line[@hits=0]")
        ]

    @memoize
    def hit_statements(self, filename):
        """
        Return a list of covered line numbers for each of the hit statements
        found for the file `filename`.
        """
        classElements = self._class_elements_by_file_name[filename]
        return [
            int(line.get("number"))
            for classElement in classElements
            for line in classElement.xpath("./lines/line[@hits>0]")
        ]

    def line_statuses(self, filename):
        """
        Return a list of tuples `(lineno, status)` of all the lines found in
        the Cobertura report for the given file `filename` where `lineno` is
        the line number and `status` is coverage status of the line which can
        be either `True` (line hit) or `False` (line miss).
        """
        line_elements = self._get_lines_by_filename(filename)
        return [
            (int(line.get("number")), line.get("hits") != "0") for line in line_elements
        ]

    def missed_lines(self, filename):
        """
        Return a list of extrapolated uncovered line numbers for the
        file `filename` according to `Cobertura.line_statuses`.
        """
        statuses = self.line_statuses(filename)
        statuses = extrapolate_coverage(statuses)
        return [lno for lno, status in statuses if status is False]

    def _raise_MissingFileSystem(self, filename):
        raise self.MissingFileSystem(
            f"Unable to read file: {filename}. "
            f"A FileSystem instance must be provided via "
            f"Cobertura(filesystem=...) to locate and read the source "
            f"content of the file."
        )

    @memoize
    def file_source(self, filename):
        """
        Return a list of namedtuple `Line` for each line of code found in the
        source file with the given `filename`.
        """
        if self.filesystem is None:
            self._raise_MissingFileSystem(filename)

        lines = []
        try:
            with self.filesystem.open(filename) as f:
                line_statuses = dict(self.line_statuses(filename))
                for lineno, source in enumerate(f, start=1):
                    line_status = line_statuses.get(lineno)
                    line = Line(lineno, source, line_status, None)
                    lines.append(line)

        except self.filesystem.FileNotFound as file_not_found:
            lines.append(Line(0, f"{file_not_found.path} not found", None, None))

        return lines

    def total_misses(self, filename=None):
        """
        Return the total number of uncovered statements for the file
        `filename`. If `filename` is not given, return the total
        number of uncovered statements for all files.
        """
        if filename is not None:
            return len(self.missed_statements(filename))

        return sum([len(self.missed_statements(filename)) for filename in self.files()])

    def total_hits(self, filename=None):
        """
        Return the total number of covered statements for the file
        `filename`. If `filename` is not given, return the total
        number of covered statements for all files.
        """
        if filename is not None:
            return len(self.hit_statements(filename))

        return sum([len(self.hit_statements(filename)) for filename in self.files()])

    def total_statements(self, filename=None):
        """
        Return the total number of statements for the file
        `filename`. If `filename` is not given, return the total
        number of statements for all files.
        """
        if filename is not None:
            return len(self._get_lines_by_filename(filename))

        return sum(
            [len(self._get_lines_by_filename(filename)) for filename in self.files()]
        )

    @memoize
    def files(self, ignore_regex=None):
        """
        Return the list of available files in the coverage report.
        """
        # maybe replace with a trie at some point? see has_file FIXME
        already_seen = set()
        filenames = []

        for el in self.xml.xpath("//class"):
            filename = el.get("filename")
            if filename in already_seen:
                continue
            already_seen.add(filename)
            filenames.append(filename)

        return (
            filenames
            if not ignore_regex
            else get_filenames_that_do_not_match_regex(filenames, ignore_regex)
        )

    def has_file(self, filename):
        """
        Return `True` if the file `filename` is present in the report, return
        `False` otherwise.
        """
        # FIXME: this will lookup a list which is slow, make it O(1)
        return filename in self.files()

    @memoize
    def source_lines(self, filename):
        """
        Return a list for source lines of file `filename`.
        """
        if self.filesystem is None:
            self._raise_MissingFileSystem(filename)

        with self.filesystem.open(filename) as f:
            return f.readlines()

    @memoize
    def packages(self):
        """
        Return the list of available packages in the coverage report.
        """
        return [el.get("name") for el in self.xml.xpath("//package")]


class CoberturaDiff:
    """
    Diff Cobertura objects.
    """

    def __init__(self, cobertura1, cobertura2):
        self.cobertura1 = cobertura1
        self.cobertura2 = cobertura2

    def has_better_coverage(self):
        """
        Return `True` if coverage of has improved, `False` otherwise.

        This does not ensure that all changes have been covered. If this is
        what you want, use `CoberturaDiff.has_all_changes_covered()` instead.
        """
        return not (
            any(self.diff_total_misses(filename) > 0 for filename in self.files())
        )

    def has_all_changes_covered(self):
        """
        Return `True` if all changes have been covered, `False` otherwise.
        """
        for filename in self.files():
            for hunk in self.file_source_hunks(filename):
                for line in hunk:
                    if line.reason is None:
                        continue  # line untouched
                    if line.status is False:
                        return False  # line not covered
        return True

    def _diff_attr(self, attr_name, filename):
        """
        Return the difference between
        `self.cobertura2.<attr_name>(filename)` and
        `self.cobertura1.<attr_name>(filename)`.

        This generic method is meant to diff the count of methods that return
        counts for a given file `filename`, e.g. `Cobertura.total_statements`,
        `Cobertura.total_misses`, ...

        The returned count may be a float.
        """
        files = [filename] if filename else self.files()

        total_count = 0.0
        for filename in files:
            if self.cobertura1.has_file(filename):
                method = getattr(self.cobertura1, attr_name)
                count1 = method(filename)
            else:
                count1 = 0.0
            method = getattr(self.cobertura2, attr_name)
            count2 = method(filename)
            total_count += count2 - count1

        return total_count

    def diff_total_statements(self, filename=None):
        return int(self._diff_attr("total_statements", filename))

    def diff_total_misses(self, filename=None):
        return int(self._diff_attr("total_misses", filename))

    def diff_total_hits(self, filename=None):
        return int(self._diff_attr("total_hits", filename))

    def diff_line_rate(self, filename=None):
        if filename is not None:
            return self._diff_attr("line_rate", filename)
        return self.cobertura2.line_rate() - self.cobertura1.line_rate()

    def diff_missed_lines(self, filename):
        """
        Return a list of 2-element tuples `(lineno, True)` for uncovered lines.
        The given file `filename` where `lineno` is a missed line number.
        """
        return [
            line.number for line in self.file_source(filename) if line.status is False
        ]

    def files(self, ignore_regex=None):
        """
        Return `self.cobertura2.files()`.
        """
        return self.cobertura2.files(ignore_regex)

    def file_source(self, filename):
        """
        Return a list of namedtuple `Line` for each line of code found in the
        given file `filename`.

        """
        if self.cobertura1.has_file(filename) and self.cobertura1.filesystem.has_file(
            filename
        ):
            lines1 = self.cobertura1.source_lines(filename)
            line_statuses1 = dict(self.cobertura1.line_statuses(filename))
        else:
            lines1 = []
            line_statuses1 = {}

        lines2 = self.cobertura2.source_lines(filename)
        line_statuses2 = dict(self.cobertura2.line_statuses(filename))

        # Build a dict of lineno2 -> lineno1
        lineno_map = reconcile_lines(lines2, lines1)

        # if we are using a single coverage file, we need to translate the
        # coverage of lines1 so that it corresponds to its real lines.
        if self.cobertura1 == self.cobertura2:
            line_statuses1 = {}
            for l2, l1 in lineno_map.items():
                line_statuses1[l1] = line_statuses2.get(l2)

        lines = []
        for lineno, source in enumerate(lines2, start=1):
            status = None
            reason = None
            if lineno not in lineno_map:
                # line was added or removed, just use whatever coverage status
                # is available as there is nothing to compare against.
                status = line_statuses2.get(lineno)
                reason = "line-edit"
            else:
                other_lineno = lineno_map[lineno]
                line_status1 = line_statuses1.get(other_lineno)
                line_status2 = line_statuses2.get(lineno)
                if line_status1 is line_status2:
                    status = None  # unchanged
                    reason = None
                elif line_status1 is True and line_status2 is False:
                    status = False  # decreased
                    reason = "cov-down"
                elif line_status1 is False and line_status2 is True:
                    status = True  # increased
                    reason = "cov-up"

            line = Line(lineno, source, status, reason)
            lines.append(line)

        return lines

    def file_source_hunks(self, filename):
        """
        Like `CoberturaDiff.file_source`, but returns a list of line hunks of
        the lines that have changed for the given file `filename`. An empty
        list means that the file has no lines that have a change in coverage
        status.
        """
        lines = self.file_source(filename)
        hunks = hunkify_lines(lines)
        return hunks



from pycobertura.utils import get_non_empty_non_commented_lines_from_file_in_ascii, get_filenames_that_do_not_match_regex

def test_read_ignore_regex_file():
    result = get_non_empty_non_commented_lines_from_file_in_ascii('tests/.testgitignore', '#')
    assert result == ['**/__pycache__','**/dummy*','*.xml']

def test_get_filenames_that_do_not_match_regex_given_file_path():
    filenames = ["tests/dummy/test_dummy.py", "tests/cobertura.xml", "tests/test_cli.py"]
    result = get_filenames_that_do_not_match_regex(filenames, "tests/.testgitignore")
    assert result == ["tests/test_cli.py"]

def test_get_filenames_that_do_not_match_regex_given_list_of_filenames():
    filenames = ["tests/dummy/test_dummy.py", "tests/cobertura.xml", "tests/test_cli.py"]
    regex_param = "^tests/dummy|(.*).xml$"
    result = get_filenames_that_do_not_match_regex(filenames, regex_param)
    assert result == ["tests/test_cli.py"]




def test_reconcile_lines__identical():
    from pycobertura.utils import reconcile_lines

    lines1 = [
        'hello',  # 1
        'world',  # 2
    ]

    lines2 = [
        'hello',  # 1
        'world',  # 2
    ]

    assert reconcile_lines(lines1, lines2) == {1: 1, 2: 2}


def test_reconcile_lines__less_to_more():
    from pycobertura.utils import reconcile_lines

    lines1 = [
        'hello',  # 1
        'world',  # 2
    ]
    lines2 = [
        'dear all',   # 1
        'hello',      # 2
        'beautiful',  # 3
        'world',      # 4
        'of',         # 5
        'pain',       # 6
    ]

    assert reconcile_lines(lines1, lines2) == {1: 2, 2: 4}


def test_reconcile_lines__more_to_less():
    from pycobertura.utils import reconcile_lines

    lines1 = [
        'dear all',   # 1
        'hello',      # 2
        'beautiful',  # 3
        'world',      # 4
        'of',         # 5
        'pain',       # 6
    ]

    lines2 = [
        'hello',  # 1
        'world',  # 2
    ]

    assert reconcile_lines(lines1, lines2) == {2: 1, 4: 2}



"""
Jinja2 filters meant to be used by templates.
"""

line_status_style = {
    True: "hit",
    False: "miss",
    None: "noop",
}


line_reason_html_icon = {
    "line-edit": "+",
    "cov-up": "&nbsp;",
    "cov-down": "&nbsp;",
    None: "&nbsp;",
}


def is_not_equal_to_dash(arg):
    return not (arg == "-")


def misses_color(arg):
    if arg.startswith("+") or arg[0].isdigit():
        return "red"
    return "green"


def line_status(line):
    return line_status_style[line.status]


def line_reason_icon(line):
    if line.status is None:
        return "&nbsp;"
    return line_reason_html_icon[line.reason]



def test_extrapolate_coverage():
    from pycobertura.utils import extrapolate_coverage
    lines_w_status = [
        (1, True),
        (4, True),
        (7, False),
        (9, False),
    ]

    assert extrapolate_coverage(lines_w_status) == [
        (1, True),
        (2, True),
        (3, True),
        (4, True),
        (5, None),
        (6, None),
        (7, False),
        (8, False),
        (9, False),
    ]



from .utils import make_cobertura


def remove_style_tag(html):
    style_pattern_start = '\n    <style>'
    style_pattern_stop = '\n    </style>'
    style_starts = html.find(style_pattern_start)
    style_stops = html.find(style_pattern_stop) + len(style_pattern_stop)
    html_nostyle = html[:style_starts] + html[style_stops:]
    return html_nostyle


def test_text_report():
    from pycobertura.reporters import TextReporter

    cobertura = make_cobertura()
    report = TextReporter(cobertura)

    assert report.generate() == """\
Filename                          Stmts    Miss  Cover    Missing
------------------------------  -------  ------  -------  ---------
Main.java                            15       0  100.00%
search/BinarySearch.java             12       1  91.67%   24
search/ISortedArraySearch.java        0       0  100.00%
search/LinearSearch.java              7       2  71.43%   19-24
TOTAL                                34       3  90.00%"""


def test_text_report__with_missing_range():
    from pycobertura.reporters import TextReporter

    cobertura = make_cobertura('tests/dummy.with-dummy2-no-cov.xml')
    report = TextReporter(cobertura)

    assert report.generate() == """\
Filename             Stmts    Miss  Cover    Missing
-----------------  -------  ------  -------  ---------
dummy/__init__.py        0       0  0.00%
dummy/dummy.py           4       0  100.00%
dummy/dummy2.py          2       2  0.00%    1-2
TOTAL                    6       2  66.67%"""


def test_text_report_delta__no_diff():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source1/coverage.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Filename      Stmts    Miss  Cover     Missing
----------  -------  ------  --------  ---------
TOTAL             0       0  +100.00%"""


def test_text_report_delta__colorize_True():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, color=True)

    assert report_delta.generate() == """\
Filename           Stmts    Miss  Cover     Missing
---------------  -------  ------  --------  ---------
dummy/dummy.py         0      \x1b[32m-2\x1b[39m  +40.00%
dummy/dummy2.py       +2      \x1b[31m+1\x1b[39m  -25.00%   \x1b[31m5\x1b[39m
dummy/dummy3.py       +2      \x1b[31m+2\x1b[39m  +100.00%  \x1b[31m1\x1b[39m, \x1b[31m2\x1b[39m
TOTAL                 +4      \x1b[31m+1\x1b[39m  +31.06%"""


def test_text_report_delta__colorize_True__with_missing_range():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, color=True)

    assert report_delta.generate() == """\
Filename           Stmts    Miss  Cover     Missing
---------------  -------  ------  --------  ---------
dummy/dummy.py         0      \x1b[32m-2\x1b[39m  +40.00%
dummy/dummy2.py       +2      \x1b[31m+1\x1b[39m  -25.00%   \x1b[31m5\x1b[39m
dummy/dummy3.py       +2      \x1b[31m+2\x1b[39m  +100.00%  \x1b[31m1\x1b[39m, \x1b[31m2\x1b[39m
TOTAL                 +4      \x1b[31m+1\x1b[39m  +31.06%"""


def test_text_report_delta__colorize_False():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, color=False)

    assert report_delta.generate() == """\
Filename           Stmts    Miss  Cover     Missing
---------------  -------  ------  --------  ---------
dummy/dummy.py         0      -2  +40.00%
dummy/dummy2.py       +2      +1  -25.00%   5
dummy/dummy3.py       +2      +2  +100.00%  1, 2
TOTAL                 +4      +1  +31.06%"""


def test_html_report():
    from pycobertura.reporters import HtmlReporter

    cobertura = make_cobertura()
    report = HtmlReporter(cobertura)
    html_output = report.generate()

    assert "normalize.css" in html_output
    assert "Skeleton V2.0" in html_output

    assert remove_style_tag(html_output) == """\
<html>
  <head>
    <title>pycobertura report</title>
    <meta charset="UTF-8">
  </head>
  <body>
    <div class="container">
      <h1>pycobertura report</h1>
      <table class="u-full-width">
        <thead>
          <tr>
            <th>Filename</th>
            <th>Stmts</th>
            <th>Miss</th>
            <th>Cover</th>
            <th>Missing</th>
          </tr>
        </thead>
        <tbody>
          <tr>
            <td><a href="#Main.java">Main.java</a></td>
            <td>15</td>
            <td>0</td>
            <td>100.00%</td>
            <td></td>
          </tr>
          <tr>
            <td><a href="#search/BinarySearch.java">search/BinarySearch.java</a></td>
            <td>12</td>
            <td>1</td>
            <td>91.67%</td>
            <td>24</td>
          </tr>
          <tr>
            <td><a href="#search/ISortedArraySearch.java">search/ISortedArraySearch.java</a></td>
            <td>0</td>
            <td>0</td>
            <td>100.00%</td>
            <td></td>
          </tr>
          <tr>
            <td><a href="#search/LinearSearch.java">search/LinearSearch.java</a></td>
            <td>7</td>
            <td>2</td>
            <td>71.43%</td>
            <td>19-24</td>
          </tr>
        </tbody>
        <tfoot>
          <tr>
            <td>TOTAL</td>
            <td>34</td>
            <td>3</td>
            <td>90.00%</td>
            <td></td>
          </tr>
        </tfoot>
      </table>
<h4 id="Main.java">Main.java</h4>
<table class="code u-max-full-width">
  <tbody>
    <tr>
      <td class="lineno">
        <pre>0 &nbsp;
</pre>
      </td>
      <td class="source">
        <pre><span class="noop">tests/Main.java not found</span></pre>
      </td>
    </tr>
  </tbody>
</table>
<h4 id="search/BinarySearch.java">search/BinarySearch.java</h4>
<table class="code u-max-full-width">
  <tbody>
    <tr>
      <td class="lineno">
        <pre>0 &nbsp;
</pre>
      </td>
      <td class="source">
        <pre><span class="noop">tests/search/BinarySearch.java not found</span></pre>
      </td>
    </tr>
  </tbody>
</table>
<h4 id="search/ISortedArraySearch.java">search/ISortedArraySearch.java</h4>
<table class="code u-max-full-width">
  <tbody>
    <tr>
      <td class="lineno">
        <pre>0 &nbsp;
</pre>
      </td>
      <td class="source">
        <pre><span class="noop">tests/search/ISortedArraySearch.java not found</span></pre>
      </td>
    </tr>
  </tbody>
</table>
<h4 id="search/LinearSearch.java">search/LinearSearch.java</h4>
<table class="code u-max-full-width">
  <tbody>
    <tr>
      <td class="lineno">
        <pre>0 &nbsp;
</pre>
      </td>
      <td class="source">
        <pre><span class="noop">tests/search/LinearSearch.java not found</span></pre>
      </td>
    </tr>
  </tbody>
</table>

    </div>
  </body>
</html>"""


def test_html_report__no_source_files_message():
    from pycobertura.reporters import HtmlReporter

    cobertura = make_cobertura()
    report = HtmlReporter(cobertura, title="test report", render_file_sources=False)
    html_output = report.generate()

    assert "normalize.css" in html_output
    assert "Skeleton V2.0" in html_output

    assert remove_style_tag(html_output) == """\
<html>
  <head>
    <title>test report</title>
    <meta charset="UTF-8">
  </head>
  <body>
    <div class="container">
      <h1>test report</h1>
      <table class="u-full-width">
        <thead>
          <tr>
            <th>Filename</th>
            <th>Stmts</th>
            <th>Miss</th>
            <th>Cover</th>
            <th>Missing</th>
          </tr>
        </thead>
        <tbody>
          <tr>
            <td>Main.java</td>
            <td>15</td>
            <td>0</td>
            <td>100.00%</td>
            <td></td>
          </tr>
          <tr>
            <td>search/BinarySearch.java</td>
            <td>12</td>
            <td>1</td>
            <td>91.67%</td>
            <td>24</td>
          </tr>
          <tr>
            <td>search/ISortedArraySearch.java</td>
            <td>0</td>
            <td>0</td>
            <td>100.00%</td>
            <td></td>
          </tr>
          <tr>
            <td>search/LinearSearch.java</td>
            <td>7</td>
            <td>2</td>
            <td>71.43%</td>
            <td>19-24</td>
          </tr>
        </tbody>
        <tfoot>
          <tr>
            <td>TOTAL</td>
            <td>34</td>
            <td>3</td>
            <td>90.00%</td>
            <td></td>
          </tr>
        </tfoot>
      </table>
<p>Rendering of source files was disabled.</p>
    </div>
  </body>
</html>"""


def test_text_report_delta__no_source():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, show_source=False)
    output = report_delta.generate()
    assert output == """\
Filename           Stmts    Miss  Cover
---------------  -------  ------  --------
dummy/dummy.py         0      -2  +40.00%
dummy/dummy2.py       +2      +1  -25.00%
dummy/dummy3.py       +2      +2  +100.00%
TOTAL                 +4      +1  +31.06%"""


def test_html_report_delta__no_source():
    from pycobertura.reporters import HtmlReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = HtmlReporterDelta(cobertura1, cobertura2, show_source=False)
    html_output = report_delta.generate()
    assert 'Missing' not in html_output
    assert '<h4 id=' not in html_output

    assert remove_style_tag(html_output) == """\
<html>
  <head>
    <title>pycobertura report</title>
    <meta charset="UTF-8">
  </head>
  <body>
    <div class="container">
      <table class="u-full-width">
        <thead>
          <tr>
            <th>Filename</th>
            <th>Stmts</th>
            <th>Miss</th>
            <th>Cover</th>
          </tr>
        </thead>
        <tbody>
          <tr>
            <td><a href="#dummy/dummy.py">dummy/dummy.py</a></td>
            <td>0</td>
            <td><span class="green">-2</span></td>
            <td>+40.00%</td>
          </tr>
          <tr>
            <td><a href="#dummy/dummy2.py">dummy/dummy2.py</a></td>
            <td>+2</td>
            <td><span class="red">+1</span></td>
            <td>-25.00%</td>
          </tr>
          <tr>
            <td><a href="#dummy/dummy3.py">dummy/dummy3.py</a></td>
            <td>+2</td>
            <td><span class="red">+2</span></td>
            <td>+100.00%</td>
          </tr>
        </tbody>
        <tfoot>
          <tr>
            <td>TOTAL</td>
            <td>+4</td>
            <td><span class="red">+1</span></td>
            <td>+31.06%</td>
          </tr>
        </tfoot>
      </table>
    </div>
  </body>
</html>"""


def test_html_report_delta():
    from pycobertura.reporters import HtmlReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = HtmlReporterDelta(cobertura1, cobertura2)
    html_output = report_delta.generate()
    assert '.red {color: red}' in html_output
    assert '.green {color: green}' in html_output
    assert "normalize.css" in html_output
    assert "Skeleton V2.0" in html_output

    assert remove_style_tag(html_output) == u"""\
<html>
  <head>
    <title>pycobertura report</title>
    <meta charset="UTF-8">
  </head>
  <body>
    <div class="container">
      <table class="u-full-width">
        <thead>
          <tr>
            <th>Filename</th>
            <th>Stmts</th>
            <th>Miss</th>
            <th>Cover</th>
            <th>Missing</th>
          </tr>
        </thead>
        <tbody>
          <tr>
            <td><a href="#dummy/dummy.py">dummy/dummy.py</a></td>
            <td>0</td>
            <td><span class="green">-2</span></td>
            <td>+40.00%</td>
            <td>
            </td>
          </tr>
          <tr>
            <td><a href="#dummy/dummy2.py">dummy/dummy2.py</a></td>
            <td>+2</td>
            <td><span class="red">+1</span></td>
            <td>-25.00%</td>
            <td><span class="red">5</span>
            </td>
          </tr>
          <tr>
            <td><a href="#dummy/dummy3.py">dummy/dummy3.py</a></td>
            <td>+2</td>
            <td><span class="red">+2</span></td>
            <td>+100.00%</td>
            <td><span class="red">1</span>, <span class="red">2</span>
            </td>
          </tr>
        </tbody>
        <tfoot>
          <tr>
            <td>TOTAL</td>
            <td>+4</td>
            <td><span class="red">+1</span></td>
            <td>+31.06%</td>
            <td></td>
          </tr>
        </tfoot>
      </table><div class="legend">
  <dl>
    <dt><code>code</code></dt><dd>coverage unchanged</dd>
    <dt class="hit"><code>code</code></dt><dd>coverage increased</dd>
    <dt class="miss"><code>code</code></dt><dd>coverage decreased</dd>
    <dt><code>+</code></dt><dd>line added or modified</dd>
  </dl>
</div>
<h4 id="dummy/dummy.py">dummy/dummy.py</h4>
<table class="code u-max-full-width">
  <tbody>
    <tr>
      <td class="lineno">
        <pre>2 &nbsp;
3 &nbsp;
4 &nbsp;
5 &nbsp;
6 +
</pre>
      </td>
      <td class="source">
        <pre><span class="noop">    pass
</span><span class="noop">
</span><span class="noop">def bar():
</span><span class="hit">    a = &#39;a&#39;
</span><span class="hit">    d = &#39;d&#39;
</span></pre>
      </td>
    </tr>
  </tbody>
</table>
<h4 id="dummy/dummy2.py">dummy/dummy2.py</h4>
<table class="code u-max-full-width">
  <tbody>
    <tr>
      <td class="lineno">
        <pre>1 &nbsp;
2 +
3 &nbsp;
4 +
5 &nbsp;
</pre>
      </td>
      <td class="source">
        <pre><span class="noop">def baz():
</span><span class="hit">    c = &#39;c&#39;
</span><span class="noop">
</span><span class="hit">def bat():
</span><span class="miss">    pass
</span></pre>
      </td>
    </tr>
  </tbody>
</table>
<h4 id="dummy/dummy3.py">dummy/dummy3.py</h4>
<table class="code u-max-full-width">
  <tbody>
    <tr>
      <td class="lineno">
        <pre>1 +
2 +
</pre>
      </td>
      <td class="source">
        <pre><span class="miss">def foobar():
</span><span class="miss">    pass  # This is a very long comment that was purposefully written so we could test how HTML rendering looks like when the boundaries of the page are reached. And here is a non-ascii char: \u015e
</span></pre>
      </td>
    </tr>
  </tbody>
</table>
    </div>
  </body>
</html>"""


def test_html_report_delta__show_missing_False():
    from pycobertura.reporters import HtmlReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = HtmlReporterDelta(cobertura1, cobertura2, show_missing=False)
    html_output = report_delta.generate()

    assert remove_style_tag(html_output) == u"""\
<html>
  <head>
    <title>pycobertura report</title>
    <meta charset="UTF-8">
  </head>
  <body>
    <div class="container">
      <table class="u-full-width">
        <thead>
          <tr>
            <th>Filename</th>
            <th>Stmts</th>
            <th>Miss</th>
            <th>Cover</th>
          </tr>
        </thead>
        <tbody>
          <tr>
            <td><a href="#dummy/dummy.py">dummy/dummy.py</a></td>
            <td>0</td>
            <td><span class="green">-2</span></td>
            <td>+40.00%</td>
          </tr>
          <tr>
            <td><a href="#dummy/dummy2.py">dummy/dummy2.py</a></td>
            <td>+2</td>
            <td><span class="red">+1</span></td>
            <td>-25.00%</td>
          </tr>
          <tr>
            <td><a href="#dummy/dummy3.py">dummy/dummy3.py</a></td>
            <td>+2</td>
            <td><span class="red">+2</span></td>
            <td>+100.00%</td>
          </tr>
        </tbody>
        <tfoot>
          <tr>
            <td>TOTAL</td>
            <td>+4</td>
            <td><span class="red">+1</span></td>
            <td>+31.06%</td>
          </tr>
        </tfoot>
      </table><div class="legend">
  <dl>
    <dt><code>code</code></dt><dd>coverage unchanged</dd>
    <dt class="hit"><code>code</code></dt><dd>coverage increased</dd>
    <dt class="miss"><code>code</code></dt><dd>coverage decreased</dd>
    <dt><code>+</code></dt><dd>line added or modified</dd>
  </dl>
</div>
<h4 id="dummy/dummy.py">dummy/dummy.py</h4>
<table class="code u-max-full-width">
  <tbody>
    <tr>
      <td class="lineno">
        <pre>2 &nbsp;
3 &nbsp;
4 &nbsp;
5 &nbsp;
6 +
</pre>
      </td>
      <td class="source">
        <pre><span class="noop">    pass
</span><span class="noop">
</span><span class="noop">def bar():
</span><span class="hit">    a = &#39;a&#39;
</span><span class="hit">    d = &#39;d&#39;
</span></pre>
      </td>
    </tr>
  </tbody>
</table>
<h4 id="dummy/dummy2.py">dummy/dummy2.py</h4>
<table class="code u-max-full-width">
  <tbody>
    <tr>
      <td class="lineno">
        <pre>1 &nbsp;
2 +
3 &nbsp;
4 +
5 &nbsp;
</pre>
      </td>
      <td class="source">
        <pre><span class="noop">def baz():
</span><span class="hit">    c = &#39;c&#39;
</span><span class="noop">
</span><span class="hit">def bat():
</span><span class="miss">    pass
</span></pre>
      </td>
    </tr>
  </tbody>
</table>
<h4 id="dummy/dummy3.py">dummy/dummy3.py</h4>
<table class="code u-max-full-width">
  <tbody>
    <tr>
      <td class="lineno">
        <pre>1 +
2 +
</pre>
      </td>
      <td class="source">
        <pre><span class="miss">def foobar():
</span><span class="miss">    pass  # This is a very long comment that was purposefully written so we could test how HTML rendering looks like when the boundaries of the page are reached. And here is a non-ascii char: \u015e
</span></pre>
      </td>
    </tr>
  </tbody>
</table>
    </div>
  </body>
</html>"""



def test_stringify_func__empty_bracket():
    from pycobertura.utils import stringify

    assert stringify([]) == ''


def test_stringify_func__1_2():
    from pycobertura.utils import stringify

    assert stringify([1, 2]) == '1-2'


def test_stringify_func__1_2_3():
    from pycobertura.utils import stringify

    assert stringify([1, 2, 3]) == '1-3'


def test_stringify_func__1_2_3_and_7():
    from pycobertura.utils import stringify

    assert stringify([1, 2, 3, 7]) == '1-3, 7'


def test_stringify_func__1_2_3_and_7_8():
    from pycobertura.utils import stringify

    assert stringify([1, 2, 3, 7, 8]) == '1-3, 7-8'


def test_stringify_func__1_2_3_and_7_8_9():
    from pycobertura.utils import stringify

    assert stringify([1, 2, 3, 7, 8, 9]) == '1-3, 7-9'


def test_stringify_func__1_and_7_8_9():
    from pycobertura.utils import stringify

    assert stringify([1, 7, 8, 9]) == '1, 7-9'


def test_stringify_func__1_2_and_7_8_9():
    from pycobertura.utils import stringify

    assert stringify([1, 2, 7, 8, 9]) == '1-2, 7-9'


def test_stringify_func__1_and_7():
    from pycobertura.utils import stringify

    assert stringify([1, 7]) == '1, 7'



import mock
import pytest
import lxml.etree as ET

from .utils import make_cobertura


def test_parse_path():
    from pycobertura import Cobertura

    xml_path = 'tests/cobertura.xml'
    with mock.patch('pycobertura.cobertura.ET.parse') as mock_parse:
        cobertura = Cobertura(xml_path)

    assert cobertura.xml is mock_parse.return_value.getroot.return_value


def test_parse_file_object():
    from pycobertura import Cobertura

    xml_path = 'tests/cobertura.xml'
    with mock.patch('pycobertura.cobertura.ET.parse') as mock_parse:
        cobertura = Cobertura(open(xml_path))

    assert cobertura.xml is mock_parse.return_value.getroot.return_value


def test_parse_string():
    from pycobertura import Cobertura

    xml_path = 'tests/cobertura.xml'
    with open(xml_path) as f:
        xml_string = f.read()
    assert ET.tostring(Cobertura(xml_path).xml) == ET.tostring(Cobertura(xml_string).xml)


def test_invalid_coverage_report():
    from pycobertura import Cobertura

    xml_path = 'non-existent.xml'
    pytest.raises(Cobertura.InvalidCoverageReport, Cobertura, xml_path)


def test_version():
    cobertura = make_cobertura()
    assert cobertura.version == '1.9'


def test_line_rate():
    cobertura = make_cobertura()
    assert cobertura.line_rate() == 0.9


def test_line_rate_by_class_file():
    cobertura = make_cobertura()
    expected_line_rates = {
        'Main.java': 1.0,
        'search/BinarySearch.java': 0.9166666666666666,
        'search/ISortedArraySearch.java': 1.0,
        'search/LinearSearch.java': 0.7142857142857143,
    }

    for filename in cobertura.files():
        assert cobertura.line_rate(filename) == \
            expected_line_rates[filename]


def test_branch_rate():
    cobertura = make_cobertura()
    assert cobertura.branch_rate() == 0.75


def test_branch_rate_by_class_file():
    cobertura = make_cobertura()
    expected_branch_rates = {
        'Main.java': 1.0,
        'search/BinarySearch.java': 0.8333333333333334,
        'search/ISortedArraySearch.java': 1.0,
        'search/LinearSearch.java': 0.6666666666666666,
    }

    for filename in cobertura.files():
        assert cobertura.branch_rate(filename) == \
            expected_branch_rates[filename]


def test_missed_statements_by_class_file():
    cobertura = make_cobertura()
    expected_missed_statements = {
        'Main.java': [],
        'search/BinarySearch.java': [24],
        'search/ISortedArraySearch.java': [],
        'search/LinearSearch.java': [19, 24],
    }

    for filename in cobertura.files():
        assert cobertura.missed_statements(filename) == \
            expected_missed_statements[filename]


def test_list_packages():
    cobertura = make_cobertura()

    packages = cobertura.packages()
    assert packages == ['', 'search']


@pytest.mark.parametrize("report, expected", [
    ('tests/cobertura-generated-by-istanbul-from-coffeescript.xml', [
        'app.coffee'
    ]),
    ('tests/cobertura.xml', [
        'Main.java',
        'search/BinarySearch.java',
        'search/ISortedArraySearch.java',
        'search/LinearSearch.java'
    ])
])
def test_list_classes(report, expected):
    cobertura = make_cobertura(xml=report)

    classes = cobertura.files()
    assert classes == expected


def test_hit_lines__by_iterating_over_classes():
    cobertura = make_cobertura()

    expected_lines = {
        'Main.java': [10, 16, 17, 18, 19, 23, 25, 26, 28, 29, 30, 31, 32, 33, 34],
        'search/BinarySearch.java': [12, 16, 18, 20, 21, 23, 25, 26, 28, 29, 31],
        'search/ISortedArraySearch.java': [],
        'search/LinearSearch.java': [9, 13, 15, 16, 17],
    }

    for filename in cobertura.files():
        assert cobertura.hit_statements(filename) == expected_lines[filename]


def test_missed_lines():
    cobertura = make_cobertura()

    expected_lines = {
        'Main.java': [],
        'search/BinarySearch.java': [24],
        'search/ISortedArraySearch.java': [],
        'search/LinearSearch.java': [19, 20, 21, 22, 23, 24],
    }

    for filename in cobertura.files():
        assert cobertura.missed_lines(filename) == expected_lines[filename]


def test_total_statements():
    cobertura = make_cobertura()
    assert cobertura.total_statements() == 34


def test_total_statements_by_class_file():
    cobertura = make_cobertura()
    expected_total_statements = {
        'Main.java': 15,
        'search/BinarySearch.java': 12,
        'search/ISortedArraySearch.java': 0,
        'search/LinearSearch.java': 7,
    }
    for filename in cobertura.files():
        assert cobertura.total_statements(filename) == \
            expected_total_statements[filename]


def test_total_misses():
    cobertura = make_cobertura()
    assert cobertura.total_misses() == 3


def test_total_misses_by_class_file():
    cobertura = make_cobertura()
    expected_total_misses = {
        'Main.java': 0,
        'search/BinarySearch.java': 1,
        'search/ISortedArraySearch.java': 0,
        'search/LinearSearch.java': 2,
    }
    for filename in cobertura.files():
        assert cobertura.total_misses(filename) == \
            expected_total_misses[filename]


def test_total_hits():
    cobertura = make_cobertura()
    assert cobertura.total_hits() == 31


def test_total_hits_by_class_file():
    cobertura = make_cobertura()
    expected_total_misses = {
        'Main.java': 15,
        'search/BinarySearch.java': 11,
        'search/ISortedArraySearch.java': 0,
        'search/LinearSearch.java': 5,
    }
    for filename in cobertura.files():
        assert cobertura.total_hits(filename) == \
            expected_total_misses[filename]


def test_class_file_source__sources_not_found():
    from pycobertura.cobertura import Line
    cobertura = make_cobertura()
    expected_sources = {
        'Main.java': [Line(0, 'tests/Main.java not found', None, None)],
        'search/BinarySearch.java': [Line(0, 'tests/search/BinarySearch.java not found', None, None)],
        'search/ISortedArraySearch.java': [Line(0, 'tests/search/ISortedArraySearch.java not found', None, None)],
        'search/LinearSearch.java': [Line(0, 'tests/search/LinearSearch.java not found', None, None)],
    }
    for filename in cobertura.files():
        assert cobertura.file_source(filename) == expected_sources[filename]


def test_line_statuses():
    cobertura = make_cobertura('tests/dummy.source1/coverage.xml')
    expected_line_statuses = {
        'dummy/__init__.py': [],
        'dummy/dummy.py': [
            (1, True),
            (2, True),
            (4, True),
            (5, False),
            (6, False),
        ],
        'dummy/dummy2.py': [
            (1, True),
            (2, True),
        ],
        'dummy/dummy4.py': [
            (1, False),
            (2, False),
            (4, False),
            (5, False),
            (6, False)
        ],
    }
    for filename in cobertura.files():
        assert cobertura.line_statuses(filename) == \
            expected_line_statuses[filename]


@pytest.mark.parametrize("report, source, source_prefix", [
    ("tests/dummy.source1/coverage.xml", None, None),
    ("tests/dummy.source1/coverage.xml", "tests/", "dummy.source1/"),
])
def test_class_file_source__sources_found(report, source, source_prefix):
    from pycobertura.cobertura import Line
    cobertura = make_cobertura(report, source=source, source_prefix=source_prefix)
    expected_sources = {
    'dummy/__init__.py': [],
        'dummy/dummy.py': [
            Line(1, 'def foo():\n', True, None),
            Line(2, '    pass\n', True, None),
            Line(3, '\n', None, None),
            Line(4, 'def bar():\n', True, None),
            Line(5, "    a = 'a'\n", False, None),
            Line(6, "    b = 'b'\n", False, None),
        ],
        'dummy/dummy2.py': [
            Line(1, 'def baz():\n', True, None),
            Line(2, '    pass\n', True, None)
        ],
        'dummy/dummy4.py': [
            Line(1, 'def barbaz():\n', False, None),
            Line(2, '    pass\n', False, None),
            Line(3, '\n', None, None),
            Line(4, 'def foobarbaz():\n', False, None),
            Line(5, '    a = 1 + 3\n', False, None),
            Line(6, '    pass\n', False, None)
        ],
    }
    for filename in cobertura.files():
        assert cobertura.file_source(filename) == \
               expected_sources[filename]


def test_class_file_source__raises_when_no_filesystem():
    from pycobertura.cobertura import Cobertura
    cobertura = Cobertura('tests/cobertura.xml')
    for filename in cobertura.files():
        pytest.raises(
            Cobertura.MissingFileSystem,
            cobertura.file_source,
            filename
        )

def test_class_source_lines__raises_when_no_filesystem():
    from pycobertura.cobertura import Cobertura
    cobertura = Cobertura('tests/cobertura.xml')
    for filename in cobertura.files():
        pytest.raises(
            Cobertura.MissingFileSystem,
            cobertura.source_lines,
            filename
        )



def test_hunkify_coverage__top():
    from pycobertura.utils import hunkify_lines
    from pycobertura.cobertura import Line

    lines = [
        Line(1, 'a', True, None),
        Line(2, 'b', True, None),
        Line(3, 'c', None, None),
        Line(4, 'd', None, None),
        Line(5, 'e', None, None),
        Line(6, 'f', None, None),
        Line(7, 'g', None, None),
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            Line(1, 'a', True, None),
            Line(2, 'b', True, None),
            Line(3, 'c', None, None),
            Line(4, 'd', None, None),
            Line(5, 'e', None, None),
        ],
    ]


def test_hunkify_coverage__bottom():
    from pycobertura.utils import hunkify_lines
    from pycobertura.cobertura import Line

    lines = [
        Line(1, 'a', None, None),
        Line(2, 'b', None, None),
        Line(3, 'c', None, None),
        Line(4, 'd', None, None),
        Line(5, 'e', None, None),
        Line(6, 'f', True, None),
        Line(7, 'g', True, None),
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            Line(3, 'c', None, None),
            Line(4, 'd', None, None),
            Line(5, 'e', None, None),
            Line(6, 'f', True, None),
            Line(7, 'g', True, None),
        ],
    ]


def test_hunkify_coverage__middle():
    from pycobertura.utils import hunkify_lines
    from pycobertura.cobertura import Line

    lines = [
        Line(1, 'a', None, None),
        Line(2, 'b', None, None),
        Line(3, 'c', None, None),
        Line(4, 'd', False, None),
        Line(5, 'e', False, None),
        Line(6, 'f', None, None),
        Line(7, 'g', None, None),
        Line(8, 'h', None, None),
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            Line(1, 'a', None, None),
            Line(2, 'b', None, None),
            Line(3, 'c', None, None),
            Line(4, 'd', False, None),
            Line(5, 'e', False, None),
            Line(6, 'f', None, None),
            Line(7, 'g', None, None),
            Line(8, 'h', None, None),
        ],
    ]


def test_hunkify_coverage__overlapping():
    from pycobertura.utils import hunkify_lines
    from pycobertura.cobertura import Line

    lines = [
        Line(1, 'a', None, None),
        Line(2, 'b', True, None),
        Line(3, 'c', True, None),
        Line(4, 'd', None, None),
        Line(5, 'e', None, None),
        Line(6, 'f', True, None),
        Line(7, 'g', True, None),
        Line(8, 'h', None, None),
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            Line(1, 'a', None, None),
            Line(2, 'b', True, None),
            Line(3, 'c', True, None),
            Line(4, 'd', None, None),
            Line(5, 'e', None, None),
            Line(6, 'f', True, None),
            Line(7, 'g', True, None),
            Line(8, 'h', None, None),
        ],
    ]


def test_hunkify_coverage__2_contiguous_hunks_w_full_context_makes_1_hunk():
    from pycobertura.utils import hunkify_lines
    from pycobertura.cobertura import Line

    lines = [
        Line(1, 'a', None, None),  # context 1
        Line(2, 'b', None, None),  # context 1
        Line(3, 'c', None, None),  # context 1
        Line(4, 'd', True, None),
        Line(5, 'e', None, None),  # context 1
        Line(6, 'f', None, None),  # context 1
        Line(7, 'g', None, None),  # context 1
        Line(8, 'h', None, None),  # context 2
        Line(9, 'i', None, None),  # context 2
        Line(10, 'j', None, None),  # context 2
        Line(11, 'k', False, None),
        Line(12, 'l', None, None),  # context 2
        Line(13, 'm', None, None),  # context 2
        Line(14, 'n', None, None),  # context 2
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            Line(1, 'a', None, None),
            Line(2, 'b', None, None),
            Line(3, 'c', None, None),
            Line(4, 'd', True, None),
            Line(5, 'e', None, None),
            Line(6, 'f', None, None),
            Line(7, 'g', None, None),
            Line(8, 'h', None, None),
            Line(9, 'i', None, None),
            Line(10, 'j', None, None),
            Line(11, 'k', False, None),
            Line(12, 'l', None, None),
            Line(13, 'm', None, None),
            Line(14, 'n', None, None),
        ],
    ]


def test_hunkify_coverage__2_distant_hunks_w_full_context_makes_2_hunk():
    from pycobertura.utils import hunkify_lines
    from pycobertura.cobertura import Line

    lines = [
        Line(1, 'a', None, None),  # context 1
        Line(2, 'b', None, None),  # context 1
        Line(3, 'c', None, None),  # context 1
        Line(4, 'd', True, None),
        Line(5, 'e', None, None),  # context 1
        Line(6, 'f', None, None),  # context 1
        Line(7, 'g', None, None),  # context 1
        Line(8, 'h', None, None),
        Line(9, 'i', None, None),  # context 2
        Line(10, 'j', None, None),  # context 2
        Line(11, 'k', None, None),  # context 2
        Line(12, 'l', False, None),
        Line(13, 'm', None, None),  # context 2
        Line(14, 'n', None, None),  # context 2
        Line(15, 'o', None, None),  # context 2
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            Line(1, 'a', None, None),
            Line(2, 'b', None, None),
            Line(3, 'c', None, None),
            Line(4, 'd', True, None),
            Line(5, 'e', None, None),
            Line(6, 'f', None, None),
            Line(7, 'g', None, None),
        ], [
            Line(9, 'i', None, None),
            Line(10, 'j', None, None),
            Line(11, 'k', None, None),
            Line(12, 'l', False, None),
            Line(13, 'm', None, None),
            Line(14, 'n', None, None),
            Line(15, 'o', None, None),
        ],
    ]


def test_hunkify_coverage__no_valid_hunks_found():
    from pycobertura.utils import hunkify_lines
    from pycobertura.cobertura import Line

    lines = [
        Line(1, 'a', None, None),
        Line(2, 'b', None, None),
        Line(3, 'c', None, None),
        Line(4, 'd', None, None),
        Line(5, 'e', None, None),
        Line(6, 'f', None, None),
        Line(7, 'g', None, None),
        Line(8, 'h', None, None),
        Line(9, 'i', None, None),
    ]

    hunks = hunkify_lines(lines)

    assert hunks == []



import os
import pytest
from click.testing import CliRunner


def test_exit_codes():
    # We shouldn't change exit codes so that clients can rely on them
    from pycobertura.cli import ExitCodes

    assert ExitCodes.OK == 0
    assert ExitCodes.EXCEPTION == 1
    assert ExitCodes.COVERAGE_WORSENED == 2
    assert ExitCodes.NOT_ALL_CHANGES_COVERED == 3


def test_show__format_default():
    from pycobertura.cli import show, ExitCodes

    runner = CliRunner()
    result = runner.invoke(
        show, ['tests/dummy.original.xml'], catch_exceptions=False
    )
    assert result.output == """\
Filename             Stmts    Miss  Cover    Missing
-----------------  -------  ------  -------  ---------
dummy/__init__.py        0       0  0.00%
dummy/dummy.py           4       2  50.00%   2, 5
TOTAL                    4       2  50.00%
"""
    assert result.exit_code == ExitCodes.OK


def test_show__format_text():
    from pycobertura.cli import show, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(
            show,
            ['tests/dummy.original.xml', opt, 'text'],
            catch_exceptions=False
        )
        assert result.output == """\
Filename             Stmts    Miss  Cover    Missing
-----------------  -------  ------  -------  ---------
dummy/__init__.py        0       0  0.00%
dummy/dummy.py           4       2  50.00%   2, 5
TOTAL                    4       2  50.00%
"""
    assert result.exit_code == ExitCodes.OK

def test_show__format_csv():
    from pycobertura.cli import show, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(
            show,
            ['tests/dummy.original.xml', opt, 'csv'],
            catch_exceptions=False
        )
        assert result.output == """\
Filename;Stmts;Miss;Cover;Missing
dummy/__init__.py;0;0;0.00%;
dummy/dummy.py;4;2;50.00%;2, 5
TOTAL;4;2;50.00%;
"""
    assert result.exit_code == ExitCodes.OK

def test_show__format_csv_delimiter_semicolon():
    from pycobertura.cli import show, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(
            show,
            ['tests/dummy.original.xml', opt, 'csv', '--delimiter', ';'],
            catch_exceptions=False
        )
        assert result.output == """\
Filename;Stmts;Miss;Cover;Missing
dummy/__init__.py;0;0;0.00%;
dummy/dummy.py;4;2;50.00%;2, 5
TOTAL;4;2;50.00%;
"""
    assert result.exit_code == ExitCodes.OK

def test_show__format_csv_delimiter_tab():
    from pycobertura.cli import show, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(
            show,
            ['tests/dummy.original.xml', opt, 'csv', '--delimiter', '\t'],
            catch_exceptions=False
        )
        assert result.output == """\
Filename\tStmts\tMiss\tCover\tMissing
dummy/__init__.py\t0\t0\t0.00%\t
dummy/dummy.py\t4\t2\t50.00%\t2, 5
TOTAL\t4\t2\t50.00%\t
"""
    assert result.exit_code == ExitCodes.OK

def test_show__format_markdown():
    from pycobertura.cli import show, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(
            show,
            ['tests/dummy.original.xml', opt, 'markdown'],
            catch_exceptions=False
        )
        assert result.output == """\
| Filename          |   Stmts |   Miss | Cover   | Missing   |
|-------------------|---------|--------|---------|-----------|
| dummy/__init__.py |       0 |      0 | 0.00%   |           |
| dummy/dummy.py    |       4 |      2 | 50.00%  | 2, 5      |
| TOTAL             |       4 |      2 | 50.00%  |           |
"""
    assert result.exit_code == ExitCodes.OK


def test_show__format_html():
    from pycobertura.cli import show, ExitCodes

    runner = CliRunner()
    result = runner.invoke(show, [
        'tests/dummy.original.xml', '--format', 'html'
    ], catch_exceptions=False)
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')
    assert result.exit_code == ExitCodes.OK


def test_show__format_json():
    from pycobertura.cli import show, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(
            show,
            ['tests/dummy.original.xml', opt, 'json'],
            catch_exceptions=False
        )
        assert result.output == """\
{
    "files": [
        {
            "Filename": "dummy/__init__.py",
            "Stmts": 0,
            "Miss": 0,
            "Cover": "0.00%",
            "Missing": ""
        },
        {
            "Filename": "dummy/dummy.py",
            "Stmts": 4,
            "Miss": 2,
            "Cover": "50.00%",
            "Missing": "2, 5"
        }
    ],
    "total": {
        "Filename": "TOTAL",
        "Stmts": 4,
        "Miss": 2,
        "Cover": "50.00%"
    }
}
"""
    assert result.exit_code == ExitCodes.OK

def test_show__format_yaml():
    from pycobertura.cli import show, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(
            show,
            ['tests/dummy.original.xml', opt, 'yaml'],
            catch_exceptions=False
        )
        assert result.output == """\
files:
- Filename: dummy/__init__.py
  Stmts: 0
  Miss: 0
  Cover: 0.00%
  Missing: ''
- Filename: dummy/dummy.py
  Stmts: 4
  Miss: 2
  Cover: 50.00%
  Missing: 2, 5
total:
  Filename: TOTAL
  Stmts: 4
  Miss: 2
  Cover: 50.00%

"""
    assert result.exit_code == ExitCodes.OK


def test_show__output_to_file():
    from pycobertura.cli import show, ExitCodes

    runner = CliRunner()
    for opt in ('-o', '--output'):
        result = runner.invoke(show, [
            'tests/cobertura.xml', opt, 'report.out'
        ], catch_exceptions=False)
        with open('report.out') as f:
            report = f.read()
        os.remove('report.out')
        assert result.output == ""
        assert report == """\
Filename                          Stmts    Miss  Cover    Missing
------------------------------  -------  ------  -------  ---------
Main.java                            15       0  100.00%
search/BinarySearch.java             12       1  91.67%   24
search/ISortedArraySearch.java        0       0  100.00%
search/LinearSearch.java              7       2  71.43%   19-24
TOTAL                                34       3  90.00%"""
    assert result.exit_code == ExitCodes.OK


def test_diff__format_default():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
    ], catch_exceptions=False)
    assert result.output == """\
Filename           Stmts    Miss  Cover     Missing
---------------  -------  ------  --------  ---------
dummy/dummy.py         0      \x1b[32m-2\x1b[39m  +40.00%
dummy/dummy2.py       +2      \x1b[31m+1\x1b[39m  -25.00%   \x1b[31m5\x1b[39m
dummy/dummy3.py       +2      \x1b[31m+2\x1b[39m  +100.00%  \x1b[31m1\x1b[39m, \x1b[31m2\x1b[39m
TOTAL                 +4      \x1b[31m+1\x1b[39m  +31.06%
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__format_text():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(diff, [
            opt, 'text',
            'tests/dummy.source1/coverage.xml',
            'tests/dummy.source2/coverage.xml',
        ], catch_exceptions=False)
        assert result.output == """\
Filename           Stmts    Miss  Cover     Missing
---------------  -------  ------  --------  ---------
dummy/dummy.py         0      \x1b[32m-2\x1b[39m  +40.00%
dummy/dummy2.py       +2      \x1b[31m+1\x1b[39m  -25.00%   \x1b[31m5\x1b[39m
dummy/dummy3.py       +2      \x1b[31m+2\x1b[39m  +100.00%  \x1b[31m1\x1b[39m, \x1b[31m2\x1b[39m
TOTAL                 +4      \x1b[31m+1\x1b[39m  +31.06%
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED

def test_diff__format_csv():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(diff, [
            opt, 'csv',
            'tests/dummy.source1/coverage.xml',
            'tests/dummy.source2/coverage.xml',
        ], catch_exceptions=False)
        assert result.output == """\
Filename;Stmts;Miss;Cover;Missing
dummy/dummy.py;0;\x1b[32m-2\x1b[39m;+40.00%;[]
dummy/dummy2.py;+2;\x1b[31m+1\x1b[39m;-25.00%;['\x1b[31m5\x1b[39m']
dummy/dummy3.py;+2;\x1b[31m+2\x1b[39m;+100.00%;['\x1b[31m1\x1b[39m', '\x1b[31m2\x1b[39m']
TOTAL;+4;\x1b[31m+1\x1b[39m;+31.06%;[]
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED

def test_diff__format_csv_delimiter_semicolon():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    for delim_opt in ('-delim', '--delimiter'):
        for opt in ('-f', '--format'):
            result = runner.invoke(diff, [
                opt, 'csv',
                'tests/dummy.source1/coverage.xml',
                'tests/dummy.source2/coverage.xml',
                delim_opt, ';',
        ], catch_exceptions=False)
        assert result.output == """\
Filename;Stmts;Miss;Cover;Missing
dummy/dummy.py;0;\x1b[32m-2\x1b[39m;+40.00%;[]
dummy/dummy2.py;+2;\x1b[31m+1\x1b[39m;-25.00%;['\x1b[31m5\x1b[39m']
dummy/dummy3.py;+2;\x1b[31m+2\x1b[39m;+100.00%;['\x1b[31m1\x1b[39m', '\x1b[31m2\x1b[39m']
TOTAL;+4;\x1b[31m+1\x1b[39m;+31.06%;[]
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED

def test_diff__format_csv_delimiter_tab():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    for delim_opt in ('-delim', '--delimiter'):
        for opt in ('-f', '--format'):
            result = runner.invoke(diff, [
                opt, 'csv',
                'tests/dummy.source1/coverage.xml',
                'tests/dummy.source2/coverage.xml',
                delim_opt, '\t'
            ], catch_exceptions=False)
        assert result.output == """\
Filename\tStmts\tMiss\tCover\tMissing
dummy/dummy.py\t0\t\x1b[32m-2\x1b[39m\t+40.00%\t[]
dummy/dummy2.py\t+2\t\x1b[31m+1\x1b[39m\t-25.00%\t['\x1b[31m5\x1b[39m']
dummy/dummy3.py\t+2\t\x1b[31m+2\x1b[39m\t+100.00%\t['\x1b[31m1\x1b[39m', '\x1b[31m2\x1b[39m']
TOTAL\t+4\t\x1b[31m+1\x1b[39m\t+31.06%\t[]
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED

def test_diff__format_markdown():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(diff, [
            opt, 'markdown',
            'tests/dummy.source1/coverage.xml',
            'tests/dummy.source2/coverage.xml',
        ], catch_exceptions=False)
    assert result.output == """\
| Filename        |   Stmts |   Miss | Cover    | Missing   |
|-----------------|---------|--------|----------|-----------|
| dummy/dummy.py  |       0 |     \x1b[32m-2\x1b[39m | +40.00%  |           |
| dummy/dummy2.py |      +2 |     \x1b[31m+1\x1b[39m | -25.00%  | \x1b[31m5\x1b[39m         |
| dummy/dummy3.py |      +2 |     \x1b[31m+2\x1b[39m | +100.00% | \x1b[31m1\x1b[39m, \x1b[31m2\x1b[39m      |
| TOTAL           |      +4 |     \x1b[31m+1\x1b[39m | +31.06%  |           |
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__format_json():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(diff, [
            opt, 'json',
            'tests/dummy.source1/coverage.xml',
            'tests/dummy.source2/coverage.xml',
        ], catch_exceptions=False)
        assert result.output == """\
{
    "files": [
        {
            "Filename": "dummy/dummy.py",
            "Stmts": "0",
            "Miss": "\u001b[32m-2\x1b[39m",
            "Cover": "+40.00%",
            "Missing": ""
        },
        {
            "Filename": "dummy/dummy2.py",
            "Stmts": "+2",
            "Miss": "\u001b[31m+1\u001b[39m",
            "Cover": "-25.00%",
            "Missing": "\u001b[31m5\u001b[39m"
        },
        {
            "Filename": "dummy/dummy3.py",
            "Stmts": "+2",
            "Miss": "\u001b[31m+2\u001b[39m",
            "Cover": "+100.00%",
            "Missing": "\u001b[31m1\u001b[39m, \u001b[31m2\u001b[39m"
        }
    ],
    "total": {
        "Filename": "TOTAL",
        "Stmts": "+4",
        "Miss": "\u001b[31m+1\u001b[39m",
        "Cover": "+31.06%"
    }
}
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED

def test_diff__format_yaml():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(diff, [
            opt, 'yaml',
            'tests/dummy.source1/coverage.xml',
            'tests/dummy.source2/coverage.xml',
        ], catch_exceptions=False)
        assert result.output == """\
files:
- Filename: dummy/dummy.py
  Stmts: '0'
  Miss: "\x1b[32m-2\x1b[39m"
  Cover: +40.00%
  Missing: ''
- Filename: dummy/dummy2.py
  Stmts: '+2'
  Miss: "\x1b[31m+1\x1b[39m"
  Cover: -25.00%
  Missing: "\x1b[31m5\x1b[39m"
- Filename: dummy/dummy3.py
  Stmts: '+2'
  Miss: "\x1b[31m+2\x1b[39m"
  Cover: +100.00%
  Missing: "\x1b[31m1\x1b[39m, \x1b[31m2\x1b[39m"
total:
  Filename: TOTAL
  Stmts: '+4'
  Miss: "\x1b[31m+1\x1b[39m"
  Cover: +31.06%

"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__output_to_file():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()

    for opt in ('-o', '--output'):
        result = runner.invoke(diff, [
            'tests/dummy.source1/coverage.xml',
            'tests/dummy.source2/coverage.xml',
            opt, 'report.out'
        ], catch_exceptions=False)
        with open('report.out') as f:
            report = f.read()
        os.remove('report.out')
        assert result.output == ""
        assert report == """\
Filename           Stmts    Miss  Cover     Missing
---------------  -------  ------  --------  ---------
dummy/dummy.py         0      -2  +40.00%
dummy/dummy2.py       +2      +1  -25.00%   5
dummy/dummy3.py       +2      +2  +100.00%  1, 2
TOTAL                 +4      +1  +31.06%"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__output_to_file__force_color():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()

    result = runner.invoke(diff, [
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
        '--color', '--output', 'report.out'
    ], catch_exceptions=False)
    with open('report.out') as f:
        report = f.read()
    os.remove('report.out')
    assert result.output == ""
    assert report == """\
Filename           Stmts    Miss  Cover     Missing
---------------  -------  ------  --------  ---------
dummy/dummy.py         0      \x1b[32m-2\x1b[39m  +40.00%
dummy/dummy2.py       +2      \x1b[31m+1\x1b[39m  -25.00%   \x1b[31m5\x1b[39m
dummy/dummy3.py       +2      \x1b[31m+2\x1b[39m  +100.00%  \x1b[31m1\x1b[39m, \x1b[31m2\x1b[39m
TOTAL                 +4      \x1b[31m+1\x1b[39m  +31.06%"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__format_text__with_color():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--color',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
    ], catch_exceptions=False)
    assert result.output == """\
Filename           Stmts    Miss  Cover     Missing
---------------  -------  ------  --------  ---------
dummy/dummy.py         0      \x1b[32m-2\x1b[39m  +40.00%
dummy/dummy2.py       +2      \x1b[31m+1\x1b[39m  -25.00%   \x1b[31m5\x1b[39m
dummy/dummy3.py       +2      \x1b[31m+2\x1b[39m  +100.00%  \x1b[31m1\x1b[39m, \x1b[31m2\x1b[39m
TOTAL                 +4      \x1b[31m+1\x1b[39m  +31.06%
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__format_text__with_no_color():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--no-color',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
    ], catch_exceptions=False)
    assert result.output == """\
Filename           Stmts    Miss  Cover     Missing
---------------  -------  ------  --------  ---------
dummy/dummy.py         0      -2  +40.00%
dummy/dummy2.py       +2      +1  -25.00%   5
dummy/dummy3.py       +2      +2  +100.00%  1, 2
TOTAL                 +4      +1  +31.06%
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED

def test_diff__format_markdown__with_color():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--color',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
        '--format',
        'markdown',
    ], catch_exceptions=False)
    assert result.output == """\
| Filename        |   Stmts |   Miss | Cover    | Missing   |
|-----------------|---------|--------|----------|-----------|
| dummy/dummy.py  |       0 |     \x1b[32m-2\x1b[39m | +40.00%  |           |
| dummy/dummy2.py |      +2 |     \x1b[31m+1\x1b[39m | -25.00%  | \x1b[31m5\x1b[39m         |
| dummy/dummy3.py |      +2 |     \x1b[31m+2\x1b[39m | +100.00% | \x1b[31m1\x1b[39m, \x1b[31m2\x1b[39m      |
| TOTAL           |      +4 |     \x1b[31m+1\x1b[39m | +31.06%  |           |
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED

def test_diff__format_markdown__with_no_color():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--no-color',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
        '--format',
        'markdown',
    ], catch_exceptions=False)
    assert result.output == """\
| Filename        |   Stmts |   Miss | Cover    | Missing   |
|-----------------|---------|--------|----------|-----------|
| dummy/dummy.py  |       0 |     -2 | +40.00%  |           |
| dummy/dummy2.py |      +2 |     +1 | -25.00%  | 5         |
| dummy/dummy3.py |      +2 |     +2 | +100.00% | 1, 2      |
| TOTAL           |      +4 |     +1 | +31.06%  |           |
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__format_json__with_color():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--color',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
        '--format',
        'json'
    ], catch_exceptions=False)
    assert result.output == """\
{
    "files": [
        {
            "Filename": "dummy/dummy.py",
            "Stmts": "0",
            "Miss": "\u001b[32m-2\x1b[39m",
            "Cover": "+40.00%",
            "Missing": ""
        },
        {
            "Filename": "dummy/dummy2.py",
            "Stmts": "+2",
            "Miss": "\u001b[31m+1\u001b[39m",
            "Cover": "-25.00%",
            "Missing": "\u001b[31m5\u001b[39m"
        },
        {
            "Filename": "dummy/dummy3.py",
            "Stmts": "+2",
            "Miss": "\u001b[31m+2\u001b[39m",
            "Cover": "+100.00%",
            "Missing": "\u001b[31m1\u001b[39m, \u001b[31m2\u001b[39m"
        }
    ],
    "total": {
        "Filename": "TOTAL",
        "Stmts": "+4",
        "Miss": "\u001b[31m+1\u001b[39m",
        "Cover": "+31.06%"
    }
}
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__format_json__with_no_color():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--no-color',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
        '--format',
        'json'
    ], catch_exceptions=False)
    assert result.output == """\
{
    "files": [
        {
            "Filename": "dummy/dummy.py",
            "Stmts": "0",
            "Miss": "-2",
            "Cover": "+40.00%",
            "Missing": ""
        },
        {
            "Filename": "dummy/dummy2.py",
            "Stmts": "+2",
            "Miss": "+1",
            "Cover": "-25.00%",
            "Missing": "5"
        },
        {
            "Filename": "dummy/dummy3.py",
            "Stmts": "+2",
            "Miss": "+2",
            "Cover": "+100.00%",
            "Missing": "1, 2"
        }
    ],
    "total": {
        "Filename": "TOTAL",
        "Stmts": "+4",
        "Miss": "+1",
        "Cover": "+31.06%"
    }
}
"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED

def test_diff__format_yaml_with_color():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(diff, [
            '--color',
            opt, 'yaml',
            'tests/dummy.source1/coverage.xml',
            'tests/dummy.source2/coverage.xml',
        ], catch_exceptions=False)
        assert result.output == """\
files:
- Filename: dummy/dummy.py
  Stmts: '0'
  Miss: "\x1b[32m-2\x1b[39m"
  Cover: +40.00%
  Missing: ''
- Filename: dummy/dummy2.py
  Stmts: '+2'
  Miss: "\x1b[31m+1\x1b[39m"
  Cover: -25.00%
  Missing: "\x1b[31m5\x1b[39m"
- Filename: dummy/dummy3.py
  Stmts: '+2'
  Miss: "\x1b[31m+2\x1b[39m"
  Cover: +100.00%
  Missing: "\x1b[31m1\x1b[39m, \x1b[31m2\x1b[39m"
total:
  Filename: TOTAL
  Stmts: '+4'
  Miss: "\x1b[31m+1\x1b[39m"
  Cover: +31.06%

"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED

def test_diff__format_yaml_with_no_color():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(diff, [
            '--no-color',
            opt, 'yaml',
            'tests/dummy.source1/coverage.xml',
            'tests/dummy.source2/coverage.xml',
        ], catch_exceptions=False)
        assert result.output == """\
files:
- Filename: dummy/dummy.py
  Stmts: '0'
  Miss: '-2'
  Cover: +40.00%
  Missing: ''
- Filename: dummy/dummy2.py
  Stmts: '+2'
  Miss: '+1'
  Cover: -25.00%
  Missing: '5'
- Filename: dummy/dummy3.py
  Stmts: '+2'
  Miss: '+2'
  Cover: +100.00%
  Missing: 1, 2
total:
  Filename: TOTAL
  Stmts: '+4'
  Miss: '+1'
  Cover: +31.06%

"""
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__format_html__no_source_on_disk():
    from pycobertura.cli import diff
    from pycobertura.filesystem import FileSystem

    runner = CliRunner()
    pytest.raises(FileSystem.FileNotFound, runner.invoke, diff, [
        '--format', 'html',
        'tests/dummy.with-dummy2-better-cov.xml',
        'tests/dummy.with-dummy2-better-and-worse.xml',
    ], catch_exceptions=False)


@pytest.mark.parametrize("source1, source2, prefix1, prefix2", [
    ("tests/", "tests/dummy", "dummy/", ""),
    ("tests/dummy", "tests/", "", "dummy/"),
    ("tests/dummy/dummy.zip", "tests/dummy/dummy.zip", "", ""),
    ("tests/dummy/dummy-with-prefix.zip", "tests/dummy", "dummy-with-prefix", ""),
    ("tests/dummy/dummy-with-prefix.zip", "tests/dummy/dummy.zip", "dummy-with-prefix", ""),
])
def test_diff__format_html__with_source_prefix(source1, source2, prefix1, prefix2):
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--format', 'html',
        '--source1', source1,
        '--source2', source2,
        '--source-prefix1', prefix1,
        '--source-prefix2', prefix2,
        'tests/dummy.with-dummy2-better-cov.xml',
        'tests/dummy.with-dummy2-better-and-worse.xml',
    ], catch_exceptions=False)
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


@pytest.mark.parametrize("source1, source2", [
    ("tests/dummy", "tests/dummy"),
    ("tests/dummy/dummy.zip", "tests/dummy/dummy.zip"),
])
def test_diff__format_html__with_source(source1, source2):
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--format', 'html',
        '--source1', source1,
        '--source2', source2,
        'tests/dummy.with-dummy2-better-cov.xml',
        'tests/dummy.with-dummy2-better-and-worse.xml',
    ], catch_exceptions=False)
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__format_html__source():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--format', 'html',
        '--source',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
    ], catch_exceptions=False)
    assert 'Missing' in result.output
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__format_html__source_is_default():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--format', 'html',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
    ], catch_exceptions=False)
    assert 'Missing' in result.output
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__format_html__no_source():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--format', 'html',
        '--no-source',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
    ], catch_exceptions=False)
    assert 'Missing' not in result.output
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__same_coverage_has_exit_status_of_zero():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source1/coverage.xml',
    ], catch_exceptions=False)
    assert result.exit_code == ExitCodes.OK


def test_diff__better_coverage_has_exit_status_of_zero():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.original.xml',
        'tests/dummy.original-full-cov.xml',  # has no uncovered lines
        '--no-source',
    ], catch_exceptions=False)
    assert result.exit_code == ExitCodes.OK


def test_diff__worse_coverage_exit_status():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.with-dummy2-no-cov.xml',
        'tests/dummy.with-dummy2-better-and-worse.xml',  # has covered AND uncovered lines
        '--no-source',
    ], catch_exceptions=False)
    assert result.exit_code == ExitCodes.COVERAGE_WORSENED


def test_diff__changes_uncovered_but_with_better_coverage_exit_status():
    from pycobertura.cli import diff, ExitCodes

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.zeroexit1/coverage.xml',
        'tests/dummy.zeroexit2/coverage.xml',  # has uncovered changes
    ], catch_exceptions=False)
    assert result.exit_code == ExitCodes.NOT_ALL_CHANGES_COVERED


def test_diff__line_status():
    from pycobertura.cli import diff

    runner = CliRunner()
    runner.invoke(diff, [
        'tests/dummy.linestatus/test1.xml',
        'tests/dummy.linestatus/test2.xml',
    ], catch_exceptions=False)
    assert True



def foobar():
    pass  # This is a very long comment that was purposefully written so we could test how HTML rendering looks like when the boundaries of the page are reached. And here is a non-ascii char: Ş



from .utils import make_cobertura


def test_diff_class_source():
    from pycobertura.cobertura import CoberturaDiff
    from pycobertura.cobertura import Line

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')
    differ = CoberturaDiff(cobertura1, cobertura2)

    expected_sources = {
        'dummy/__init__.py': [],
        'dummy/dummy.py': [
            Line(1, u'def foo():\n', None, None),
            Line(2, u'    pass\n', None, None),
            Line(3, u'\n', None, None),
            Line(4, u'def bar():\n', None, None),
            Line(5, u"    a = 'a'\n", True, 'cov-up'),
            Line(6, u"    d = 'd'\n", True, 'line-edit')
        ],
        'dummy/dummy2.py': [
            Line(1, u'def baz():\n', None, None),
            Line(2, u"    c = 'c'\n", True, 'line-edit'),
            Line(3, u'\n', None, 'line-edit'),
            Line(4, u'def bat():\n', True, 'line-edit'),
            Line(5, u'    pass\n', False, 'cov-down')
        ],
        'dummy/dummy3.py': [
            Line(1, u'def foobar():\n', False, 'line-edit'),
            Line(2, u'    pass  # This is a very long comment that was purposefully written so we could test how HTML rendering looks like when the boundaries of the page are reached. And here is a non-ascii char: \u015e\n', False, 'line-edit')
        ],
    }

    for filename in cobertura2.files():
        assert differ.file_source(filename) == \
               expected_sources[filename]


def test_diff_total_misses():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')
    differ = CoberturaDiff(cobertura1, cobertura2)

    assert differ.diff_total_misses() == 1


def test_diff_total_misses_by_class_file():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')
    differ = CoberturaDiff(cobertura1, cobertura2)

    expected_sources = {
        'dummy/__init__.py': 0,
        'dummy/dummy.py': -2,
        'dummy/dummy2.py': 1,
        'dummy/dummy3.py': 2,
    }

    for filename in cobertura2.files():
        assert differ.diff_total_misses(filename) == \
            expected_sources[filename]


def test_diff_line_rate():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')
    differ = CoberturaDiff(cobertura1, cobertura2)

    assert differ.diff_line_rate() == 0.31059999999999993


def test_diff_line_rate_by_class_file():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')
    differ = CoberturaDiff(cobertura1, cobertura2)

    expected_sources = {
        'dummy/__init__.py': 0,
        'dummy/dummy.py': 0.4,
        'dummy/dummy2.py': -0.25,
        'dummy/dummy3.py': 0.0,
    }

    for filename in cobertura2.files():
        assert differ.diff_line_rate(filename) == \
            expected_sources[filename]


def test_diff_same_report_different_source_dirs():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura('tests/dummy.uncovered.addcode/coverage.xml', source='tests/dummy.uncovered/dummy/')
    cobertura2 = make_cobertura('tests/dummy.uncovered.addcode/coverage.xml', source='tests/dummy.uncovered.addcode/dummy/')
    differ = CoberturaDiff(cobertura1, cobertura2)

    assert differ.diff_missed_lines('dummy.py') == [3]


def test_diff_total_hits():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')
    differ = CoberturaDiff(cobertura1, cobertura2)

    assert differ.diff_total_hits() == 3


def test_diff_total_hits_by_class_file():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')
    differ = CoberturaDiff(cobertura1, cobertura2)

    expected_total_hits = {
        'dummy/__init__.py': 0,
        'dummy/dummy.py': 2,
        'dummy/dummy2.py': 1,
        'dummy/dummy3.py': 0,
    }

    for filename in cobertura2.files():
        assert differ.diff_total_hits(filename) == \
            expected_total_hits[filename]


def test_diff__has_all_changes_covered__some_changed_code_is_still_uncovered():
    from pycobertura.cobertura import Cobertura, CoberturaDiff

    cobertura1 = make_cobertura('tests/dummy.zeroexit1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.zeroexit2/coverage.xml')

    differ = CoberturaDiff(cobertura1, cobertura2)
    assert differ.has_all_changes_covered() is False


def test_diff__has_better_coverage():
    from pycobertura.cobertura import Cobertura, CoberturaDiff

    cobertura1 = Cobertura('tests/dummy.zeroexit1/coverage.xml')
    cobertura2 = Cobertura('tests/dummy.zeroexit2/coverage.xml')
    differ = CoberturaDiff(cobertura1, cobertura2)
    assert differ.has_better_coverage() is True


def test_diff__has_not_better_coverage():
    from pycobertura.cobertura import Cobertura, CoberturaDiff

    cobertura1 = Cobertura('tests/dummy.zeroexit2/coverage.xml')
    cobertura2 = Cobertura('tests/dummy.zeroexit1/coverage.xml')
    differ = CoberturaDiff(cobertura1, cobertura2)
    assert differ.has_better_coverage() is False



import click

from pycobertura.cobertura import Cobertura
from pycobertura.reporters import (
    HtmlReporter,
    TextReporter,
    CsvReporter,
    MarkdownReporter,
    JsonReporter,
    YamlReporter,
    HtmlReporterDelta,
    TextReporterDelta,
    CsvReporterDelta,
    MarkdownReporterDelta,
    JsonReporterDelta,
    YamlReporterDelta,
)
from pycobertura.filesystem import filesystem_factory
from pycobertura.utils import get_dir_from_file_path

pycobertura = click.Group()


reporters = {
    "html": HtmlReporter,
    "text": TextReporter,
    "csv": CsvReporter,
    "markdown": MarkdownReporter,
    "json": JsonReporter,
    "yaml": YamlReporter,
}


class ExitCodes:
    OK = 0
    EXCEPTION = 1
    COVERAGE_WORSENED = 2
    NOT_ALL_CHANGES_COVERED = 3


def get_exit_code(differ, source):
    # Compute the non-zero exit code. This is a 2-step process which involves
    # checking whether code coverage is any better first and then check if all
    # changes are covered (stricter) which can only be done if the source code
    # is available (and enabled via the --source option).

    if not differ.has_better_coverage():
        return ExitCodes.COVERAGE_WORSENED

    if source:
        if differ.has_all_changes_covered():
            return ExitCodes.OK
        else:
            return ExitCodes.NOT_ALL_CHANGES_COVERED
    else:
        return ExitCodes.OK


@pycobertura.command()
@click.argument("cobertura_file")
@click.option(
    "--ignore-regex",
    default=None,
    type=str,
    help="Regex for which files to ignore in output\n\t",
)
@click.option("-f", "--format", default="text", type=click.Choice(list(reporters)))
@click.option(
    "-delim",
    "--delimiter",
    default=";",
    type=str,
    help="Delimiter for csv format, e.g. ,;\n\t",
)
@click.option(
    "-o",
    "--output",
    metavar="<file>",
    type=click.File("wb"),
    help="Write output to <file> instead of stdout.",
)
@click.option(
    "-s",
    "--source",
    metavar="<source-dir-or-zip>",
    help="Provide path to source code directory for HTML output. The path can "
    "also be a zip archive instead of a directory.",
)
@click.option(
    "-p",
    "--source-prefix",
    metavar="<dir-prefix>",
    help="For every file found in the coverage report, it will use this "
    "prefix to lookup files on disk. This is especially useful when "
    "the --source is a zip archive and the files were zipped under "
    "a directory prefix that is not part of the source.",
)
def show(
    cobertura_file, ignore_regex, format, delimiter, output, source, source_prefix
):
    """show coverage summary of a Cobertura report"""

    if not source:
        source = get_dir_from_file_path(cobertura_file)

    cobertura = Cobertura(
        cobertura_file,
        filesystem=filesystem_factory(source, source_prefix=source_prefix),
    )
    Reporter = reporters[format]
    reporter = Reporter(cobertura, ignore_regex)

    if format == "csv":
        report = reporter.generate(delimiter)
    else:
        report = reporter.generate()

    if not isinstance(report, bytes):
        report = report.encode("utf-8")

    isatty = True if output is None else output.isatty()
    click.echo(report, file=output, nl=isatty)


delta_reporters = {
    "text": TextReporterDelta,
    "csv": CsvReporterDelta,
    "markdown": MarkdownReporterDelta,
    "html": HtmlReporterDelta,
    "json": JsonReporterDelta,
    "yaml": YamlReporterDelta,
}


@pycobertura.command(
    help="""\
The diff command compares and shows the changes between two Cobertura reports.

NOTE: Reporting missing lines or showing the source code with the diff command
can only be accurately computed if the versions of the source code used to
generate each of the coverage reports is accessible. By default, the source
will read from the Cobertura report and resolved relatively from the report's
location. If the source is not accessible from the report's location, the
options `--source1` and `--source2` are necessary to point to the source code
directories (or zip archives). If the source is not available at all, pass
`--no-source` but missing lines and source code will not be reported.
"""
)
@click.argument("cobertura_file1")
@click.argument("cobertura_file2")
@click.option(
    "--ignore-regex",
    default=None,
    type=str,
    help="Regex for which files to ignore in output\n\t",
)
@click.option(
    "-delim",
    "--delimiter",
    default=";",
    type=str,
    help="Delimiter for csv format, e.g. ,;\n\t",
)
@click.option(
    "--color/--no-color",
    default=None,
    help="Colorize the output. By default, pycobertura emits color codes only "
    "when standard output is connected to a terminal. This has no effect "
    "with the HTML output format.",
)
@click.option(
    "-f", "--format", default="text", type=click.Choice(list(delta_reporters))
)
@click.option(
    "-o",
    "--output",
    metavar="<file>",
    type=click.File("wb"),
    help="Write output to <file> instead of stdout.",
)
@click.option(
    "-s1",
    "--source1",
    metavar="<source-dir1-or-zip-archive>",
    help="Provide path to source code directory or zip archive of first "
    "Cobertura report. This is necessary if the filename path defined "
    "in the report is not accessible from the location of the report.",
)
@click.option(
    "-s2",
    "--source2",
    metavar="<source-dir2-or-zip-archive>",
    help="Like --source1 but for the second coverage report of the diff.",
)
@click.option(
    "-p1",
    "--source-prefix1",
    metavar="<dir-prefix1>",
    help="For every file found in the coverage report, it will use this "
    "prefix to lookup files on disk. This is especially useful when "
    "the --source1 is a zip archive and the files were zipped under "
    "a directory prefix that is not part of the source",
)
@click.option(
    "-p2",
    "--source-prefix2",
    metavar="<dir-prefix2>",
    help="Like --source-prefix1, but for applies for --source2.",
)
@click.option(
    "--source/--no-source",
    default=True,
    help="Show missing lines and source code. When enabled (default), this "
    "option requires access to the source code that was used to generate "
    "both Cobertura reports (see --source1 and --source2). When "
    "`--no-source` is passed, missing lines and the source code will "
    "not be displayed.",
)
def diff(
    cobertura_file1,
    cobertura_file2,
    ignore_regex,
    delimiter,
    color,
    format,
    output,
    source1,
    source2,
    source_prefix1,
    source_prefix2,
    source,
):
    """compare coverage of two Cobertura reports"""
    # Assume that the source is located in the same directory as the provided
    # coverage files if no source directories are provided.
    if not source1:
        source1 = get_dir_from_file_path(cobertura_file1)

    if not source2:
        source2 = get_dir_from_file_path(cobertura_file2)

    filesystem1 = filesystem_factory(source1, source_prefix=source_prefix1)
    cobertura1 = Cobertura(cobertura_file1, filesystem=filesystem1)

    filesystem2 = filesystem_factory(source2, source_prefix=source_prefix2)
    cobertura2 = Cobertura(cobertura_file2, filesystem=filesystem2)

    Reporter = delta_reporters[format]
    reporter_args = [cobertura1, cobertura2, ignore_regex]
    reporter_kwargs = {"show_source": source}

    isatty = True if output is None else output.isatty()

    if format in {"text", "csv", "json", "markdown", "yaml"}:
        color = isatty if color is None else color is True
        reporter_kwargs["color"] = color

    reporter = Reporter(*reporter_args, **reporter_kwargs)
    if format == "csv":
        report = reporter.generate(delimiter)
    else:
        report = reporter.generate()

    if not isinstance(report, bytes):
        report = report.encode("utf-8")

    click.echo(report, file=output, nl=isatty, color=color)

    exit_code = get_exit_code(reporter.differ, source)
    raise SystemExit(exit_code)



from jinja2 import Environment, PackageLoader
from pycobertura.cobertura import CoberturaDiff
from pycobertura.utils import green, red, stringify
from pycobertura.templates import filters
from tabulate import tabulate
from ruamel import yaml
import json
import io


env = Environment(loader=PackageLoader("pycobertura", "templates"))
env.filters["line_status"] = filters.line_status
env.filters["line_reason"] = filters.line_reason_icon
env.filters["is_not_equal_to_dash"] = filters.is_not_equal_to_dash
env.filters["misses_color"] = filters.misses_color

headers_with_missing = ["Filename", "Stmts", "Miss", "Cover", "Missing"]
headers_without_missing = ["Filename", "Stmts", "Miss", "Cover"]


class Reporter:
    def __init__(self, cobertura, ignore_regex=None):
        self.cobertura = cobertura
        self.ignore_regex = ignore_regex

    @staticmethod
    def format_line_rate(line_rate):
        return f"{line_rate:.2%}"

    def get_report_lines(self):
        filenames = self.cobertura.files(ignore_regex=self.ignore_regex)
        lines = {
            "Filename": filenames.copy(),
            "Stmts": [
                self.cobertura.total_statements(filename) for filename in filenames
            ],
            "Miss": [self.cobertura.total_misses(filename) for filename in filenames],
            "Cover": [
                self.format_line_rate(self.cobertura.line_rate(filename))
                for filename in filenames
            ],
            "Missing": [
                stringify(self.cobertura.missed_lines(filename))
                for filename in filenames
            ],
        }
        lines["Filename"].append("TOTAL")
        lines["Stmts"] += [self.cobertura.total_statements()]
        lines["Miss"] += [self.cobertura.total_misses()]
        lines["Cover"] += [self.format_line_rate(self.cobertura.line_rate())]
        lines["Missing"].append("")

        return lines

    def per_file_stats(self, file_stats_dict):
        """
        Returns dict with keys `files` and `total` that contain coverage
        statistics per file and overall, respectively.

        {
            "files": [
                {
                    "Filename": "dummy/__init__.py",
                    "Stmts": 0,
                    "Miss": 0,
                    "Cover": "0.00%",
                    "Missing": "9"
                },
                ...
            ],
            "total": {
                "Filename": "TOTAL",
                "Stmts": "...",
                "Miss": "...",
                "Cover": "..."
            }
        }
        """
        file_stats_dict_items = file_stats_dict.items()
        number_of_files = len(self.cobertura.files()) + 1
        file_stats_list = [
            {
                header_name: header_value[file_index]
                for header_name, header_value in file_stats_dict_items
            }
            for file_index in range(number_of_files)
        ]
        file_stats_list[-1].pop("Missing", 'Key "Missing" not found')
        return {
            "files": file_stats_list[:-1],
            "total": file_stats_list[-1],
        }


class TextReporter(Reporter):
    def generate(self):
        lines = self.get_report_lines()
        return tabulate(lines, headers=headers_with_missing)


class CsvReporter(Reporter):
    def generate(self, delimiter):
        lines = self.get_report_lines()
        list_of_lines = [headers_with_missing]
        list_of_lines.extend(
            [[f"{item}" for item in row] for row in zip(*lines.values())]
        )

        # Explanation here:
        # https://stackoverflow.com/a/55889036/9698518
        delimiter = delimiter.encode().decode("unicode_escape")

        return "\n".join([delimiter.join(line) for line in list_of_lines])


class MarkdownReporter(Reporter):
    def generate(self):
        lines = self.get_report_lines()
        return tabulate(lines, headers=headers_with_missing, tablefmt="github")


class JsonReporter(Reporter):
    def generate(self):
        lines = self.get_report_lines()
        stats_dict = self.per_file_stats(lines)
        return json.dumps(stats_dict, indent=4)


class YamlReporter(Reporter):
    def generate(self):
        lines = self.get_report_lines()
        stats_dict = self.per_file_stats(lines)
        # need to write to a buffer as yml packages are using a streaming interface
        buf = io.BytesIO()
        yaml.YAML().dump(stats_dict, buf)
        return buf.getvalue()


class HtmlReporter(Reporter):
    def __init__(self, *args, **kwargs):
        self.title = kwargs.pop("title", "pycobertura report")
        self.render_file_sources = kwargs.pop("render_file_sources", True)
        self.no_file_sources_message = kwargs.pop(
            "no_file_sources_message", "Rendering of source files was disabled."
        )
        super(HtmlReporter, self).__init__(*args, **kwargs)

    def generate(self):
        lines = self.get_report_lines()

        sources = []
        if self.render_file_sources:
            sources = [
                (filename, self.cobertura.file_source(filename))
                for filename in self.cobertura.files()
            ]

        template = env.get_template("html.jinja2")
        rows = {k: v[:-1] for k, v in lines.items()}
        footer = {k: v[-1] for k, v in lines.items()}

        return template.render(
            title=self.title,
            lines=rows,
            footer=footer,
            sources=sources,
            no_file_sources_message=self.no_file_sources_message,
        )


class DeltaReporter:
    def __init__(
        self,
        cobertura1,
        cobertura2,
        ignore_regex=None,
        show_source=True,
        *args,
        **kwargs,
    ):
        self.differ = CoberturaDiff(cobertura1, cobertura2)
        self.show_source = show_source
        self.color = kwargs.pop("color", False)
        self.ignore_regex = ignore_regex

    def format_line_rate(self, line_rate):
        return f"{line_rate:+.2%}" if line_rate else "+100.00%"

    def format_total_statements(self, total_statements):
        return f"{total_statements:+d}" if total_statements else "0"

    @staticmethod
    def format_missed_lines(missed_lines):
        return [f"{lno:d}" for lno in missed_lines]

    @staticmethod
    def determine_ANSI_color_code_function_of_number(number):
        if number.startswith("+") or number[0].isdigit():
            return red
        return green

    @classmethod
    def convert_signed_number_to_ANSI_color_coded_signed_number(cls, number):
        color_function = cls.determine_ANSI_color_code_function_of_number(number)
        return color_function(number)

    def color_number(self, numbers):
        if numbers and self.color and isinstance(numbers, str):
            return self.convert_signed_number_to_ANSI_color_coded_signed_number(numbers)
        if numbers and self.color and not isinstance(numbers, str):
            return ", ".join(
                [
                    self.convert_signed_number_to_ANSI_color_coded_signed_number(number)
                    for number in numbers
                ]
            )
        return numbers if isinstance(numbers, str) else ", ".join(numbers)

    def format_total_misses(self, total_misses):
        return self.color_number(f"{total_misses:+d}") if total_misses else "0"

    def get_report_lines(self):
        filenames = self.differ.files(ignore_regex=self.ignore_regex)
        diff_total_stmts = [
            self.differ.diff_total_statements(filename) for filename in filenames
        ]

        diff_total_miss = [
            self.differ.diff_total_misses(filename) for filename in filenames
        ]

        diff_total_cover = [
            self.differ.diff_line_rate(filename) for filename in filenames
        ]

        indexes_of_files_with_changes = [
            i
            for i in range(len(self.differ.files()))
            if any(
                (
                    diff_total_stmts[i],
                    diff_total_miss[i],
                    diff_total_cover[i],
                )
            )
        ]

        filenames_of_files_with_changes = [
            self.differ.files()[i] for i in indexes_of_files_with_changes
        ]

        lines = {
            "Filename": filenames_of_files_with_changes,
            "Stmts": [
                self.format_total_statements(diff_total_stmts[i])
                for i in indexes_of_files_with_changes
            ],
            "Miss": [
                self.format_total_misses(diff_total_miss[i])
                for i in indexes_of_files_with_changes
            ],
            "Cover": [
                self.format_line_rate(diff_total_cover[i])
                for i in indexes_of_files_with_changes
            ],
        }

        lines["Filename"].append("TOTAL")
        lines["Stmts"] += [
            self.format_total_statements(self.differ.diff_total_statements())
        ]
        lines["Miss"] += [self.format_total_misses(self.differ.diff_total_misses())]
        lines["Cover"] += [self.format_line_rate(self.differ.diff_line_rate())]

        if self.show_source:
            diff_total_missing = [
                self.differ.diff_missed_lines(filename)
                for filename in self.differ.files()
            ]
            lines["Missing"] = [
                self.format_missed_lines(diff_total_missing[i])
                for i in indexes_of_files_with_changes
            ]
            lines["Missing"].append("")

        return lines

    def per_file_stats(self, file_stats_dict):
        file_stats_dict_items = file_stats_dict.items()
        number_of_files = len(self.differ.files())
        file_stats_list = [
            {
                header_name: header_value[file_index]
                for header_name, header_value in file_stats_dict_items
            }
            for file_index in range(number_of_files)
        ]
        file_stats_list[-1].pop("Missing", 'Key "Missing" not found')
        return {
            "files": file_stats_list[:-1],
            "total": file_stats_list[-1],
        }


class TextReporterDelta(DeltaReporter):
    def generate(self):
        lines = self.get_report_lines()
        headers = headers_without_missing

        if self.show_source:
            missed_lines_colored = [
                self.color_number(line) for line in lines["Missing"]
            ]
            lines["Missing"] = missed_lines_colored
            headers = headers_with_missing
        return tabulate(lines, headers=headers)


class CsvReporterDelta(DeltaReporter):
    def generate(self, delimiter):
        lines = self.get_report_lines()

        # lines_values: List of lines dictionary values arranged in
        # tuples of Table row values
        lines_values = list(zip(*lines.values()))

        # Stringify every item in Table row values without using the Missing column
        # and store in the list list_of_lines
        list_of_lines = [headers_without_missing]
        list_of_lines.extend([[f"{item}" for item in row[:-1]] for row in lines_values])

        if self.show_source:
            # Add the Missing header to list_of_lines first inner list
            # This is a direct assignment to avoid appending an additional "Missing"
            # header in every iteration of the tests which would fail them
            list_of_lines[0] = headers_with_missing
            # Add to every list inside the list_of_lines the Missing column value
            for line_index, missing_line in enumerate(lines["Missing"]):
                # for colors, explanation see here:
                # https://stackoverflow.com/a/61273717/9698518
                list_of_lines[line_index + 1] += [
                    f"{[self.color_number(number) for number in missing_line]}".encode(
                        "utf-8"
                    ).decode("unicode_escape")
                ]

        # Explanation here:
        # https://stackoverflow.com/a/55889036/9698518
        delimiter = delimiter.encode().decode("unicode_escape")

        return "\n".join([delimiter.join(line) for line in list_of_lines])


class MarkdownReporterDelta(DeltaReporter):
    def generate(self):
        lines = self.get_report_lines()
        headers = headers_without_missing

        if self.show_source:
            missed_lines_colored = [
                self.color_number(line) for line in lines["Missing"]
            ]
            lines["Missing"] = missed_lines_colored
            headers = headers_with_missing
        return tabulate(lines, headers=headers, tablefmt="github")


class JsonReporterDelta(DeltaReporter):
    def generate(self):
        lines = self.get_report_lines()

        if self.show_source:
            missed_lines_colored = [
                self.color_number(line) for line in lines["Missing"]
            ]
            lines["Missing"] = missed_lines_colored

        stats_dict = self.per_file_stats(lines)

        json_string = json.dumps(stats_dict, indent=4)

        # for colors, explanation see here:
        # https://stackoverflow.com/a/61273717/9698518
        colored_json_string = json_string.encode("utf-8").decode("unicode_escape")

        return colored_json_string


class YamlReporterDelta(DeltaReporter):
    def generate(self):
        lines = self.get_report_lines()

        if self.show_source:
            missed_lines_colored = [
                self.color_number(line) for line in lines["Missing"]
            ]
            lines["Missing"] = missed_lines_colored

        stats_dict = self.per_file_stats(lines)
        # need to write to a buffer as yml packages are using a streaming interface
        buf = io.BytesIO()
        yaml.YAML().dump(stats_dict, buf)
        # need to replace \e escape sequence with \x1b,
        # because only the latter is supported, see also
        # the Python docs for supported formats:
        # https://docs.python.org/3/reference/lexical_analysis.html#string-and-bytes-literals
        yaml_string = buf.getvalue().replace(rb"\e", b"\x1b")
        return yaml_string


class HtmlReporterDelta(DeltaReporter):
    def __init__(self, *args, **kwargs):
        """
        Takes the same arguments as `TextReporterDelta` but also takes the keyword
        argument `show_missing` which can be set to True or False to set whether
        or not the generated report should contain a listing of missing lines in
        the summary table.
        """
        self.show_missing = kwargs.pop("show_missing", True)
        super(HtmlReporterDelta, self).__init__(*args, **kwargs)

    def generate(self):
        lines = self.get_report_lines()
        template = env.get_template("html-delta.jinja2")

        rows = {k: v[:-1] for k, v in lines.items()}
        footer = {k: v[-1] for k, v in lines.items()}

        render_kwargs = {
            "lines": rows,
            "footer": footer,
            "show_missing": self.show_missing,
            "show_source": self.show_source,
        }

        if self.show_source is True:
            render_kwargs["sources"] = []
            for filename in self.differ.files():
                differ_file_source_hunks = self.differ.file_source_hunks(filename)
                if differ_file_source_hunks:
                    render_kwargs["sources"].append(
                        (filename, differ_file_source_hunks)
                    )

        return template.render(**render_kwargs)
